/** union/find with height rule and path compaction */

package applications;

public class FastUnionFindWithHeightRule
{
   // top-level nested class
   private static class Node
   {
      // data members
      int parent;     // pointer to parent in tree
      boolean root;   // true iff node is a root
   
      // constructor
      private Node()
      {
         root = true;   // this node is a root
         parent = 1;    // node is the root of a tree whose height is 1
      }
   }

   // data member of FastUnionFind
   Node [] node;

   /** initialize n trees, one element per tree/class/set */
   public FastUnionFindWithHeightRule(int n)
   {
      // allocate an array for the nodes
      node = new Node [n + 1];

      // now allocate the nodes
      for (int e = 1; e <= n; e++)
         node[e] = new Node();
   }
   
   /** @return root of the tree containing element e */
   public int find(int e)
   {
      // j will eventually be the root
      int j = e;
      while (!node[j].root)
         j = node[j].parent;
      
      // compact path from e to j
      int f = e;              // start at e
      while (f != j)
      {// f is not root
         int pf = node[f].parent;
         node[f].parent = j;  // move f to level 2
         f = pf;              // f moves to old parent 
      }
   
      return j;
   }
   
   /** combine trees with roots i and j */
   public void union(int i, int j)
   {// Use the height rule.
      if (node[i].parent < node[j].parent)
      {// i becomes subtree of j
         node[i].root = false;
         node[i].parent = j;
      }
      else
      {// j becomes subtree of i
         if (node[i].parent == node[j].parent)
            // height of i increases by 1
            node[i].parent++;
         node[j].root = false;
         node[j].parent = i;
      }
   }

   /** test program */
   public static void main(String [] args)
   {
      FastUnionFindWithHeightRule x = new FastUnionFindWithHeightRule(10);
      x.union(1, 2);
      x.union(3, 4);
      x.union(1, 3);
      System.out.println("find(1) = " + x.find(1) +
                         " find(2) = " + x.find(2));
      System.out.println("find(3) = " + x.find(3) +
                         " find(4) = " + x.find(4));
      System.out.println("find(5) = " + x.find(5) +
                         " find(6) = " + x.find(6));
   }
}
