
/** bounded-performance 0/1 knapsack heuristic with k = 1 */

package applications;

import dataStructures.*;

public class GreedyKnapsack2
{
   static class Element implements Comparable
   {
      // data members
      int id;   // object identifier
      double d; // profit density

      // constructor
      private Element(int theID, double theDensity)
      {
         id = theID;
         d = theDensity;
      }

     // method of Comparable
     /** @return true iff this < x */
     public int compareTo(Object x)
     {
        double xD = ((Element) x).d;
        if (d < xD)
           return -1;
        if (d == xD)
           return 0;
        return 1;
      }
  
      /** @return true iff this == x */
      public boolean equals(Object x)
         {return d == ((Element) x).d;}
   }

   /** bounded performance greedy heuristic for 0/1 knapsack problem, k = 1
     * set x[i] = 1 iff object i included in knapsack, 1 <= i < p.length
     * @param p[1:p.length-1] array of profits
     * @param w[1:w.length-1] array of weights
     * @param c knapsack capacity
     * @return value of greedy knapsack filling */
   public static double greedyKnapsack(double [] p, double [] w, double c,
                                       int [] x)
   {
      // define an element array to be sorted by profit density
      Element [] q = new Element [p.length];
      for (int i = 1; i < p.length; i++)
         // array of profit densities
         q[i] = new Element(i, p[i] / w[i]);
   
      // sort by density
      HeapSort.heapSort(q);
   
      int bestFix = 0;
      double maxProfit = 0;            // max profit of any solution so far
   
      // first compute value of best filling
      for (int fix = 1; fix < p.length; fix++)
      {// put object q[fix].id into the knapsack if it fits
         int fixId = q[fix].id;
         double remainingCapacity = c - w[fixId];
         if (remainingCapacity >= 0)
         {// object fixID fits
            double currentProfit = p[fixId];
      
            // fill remaining capacity in
            // nonincreasing order of density
            for (int i = p.length - 1; i > 0; i--)
            {
               int id = q[i].id;
               if (id != fixId && w[id] <= remainingCapacity)
               {// object id fits
                  remainingCapacity -= w[id];
                  currentProfit += p[id];
               }
            }
   
            if (currentProfit > maxProfit)
            {// found a better solution
               maxProfit = currentProfit;
               bestFix = fix;
            }
         }
      }

      // now compute the x[] values
      // by redoing the computation with fix = bestFix
      int fixId = q[bestFix].id;
      double remainingCapacity = c - w[fixId];
      x[fixId] = 1;
   
      // fill remaining capacity in nonincreasing order of density
      for (int i = p.length - 1; i > 0; i--)
      {
         int id = q[i].id;
         if (id != fixId)
            if (w[id] <= remainingCapacity)
            {// object id fits
               remainingCapacity -= w[id];
               x[id] = 1;
            }
         else
            x[id] = 0;
      }
   
      return maxProfit;
   }
   
   /** test greedyKnapsack */
   public static void main(String [] args)
   {
      double [] p = {0, 4, 3, 10, 1, 6};
      double [] w = {0, 2, 3, 12, 8, 4};
      int [] x = new int [6];
      int n = 5;
      double c = 20;
      System.out.print("Best value is ");
      System.out.println(greedyKnapsack(p, w, c, x));
      System.out.print("x vector is ");
      for (int i = 1; i <= n; i++)
         System.out.print(x[i] + "  ");
      System.out.println();
   }
}
