
/** bounded-performance 0/1 knapsack heuristic with k = 2 */

package applications;

import dataStructures.*;

public class GreedyKnapsack3
{
   static class Element implements Comparable
   {
      // data members
      int id;   // object identifier
      double d; // profit density

      // constructor
      private Element(int theID, double theDensity)
      {
         id = theID;
         d = theDensity;
      }

     // method of Comparable
     /** @return true iff this < x */
     public int compareTo(Object x)
     {
        double xD = ((Element) x).d;
        if (d < xD)
           return -1;
        if (d == xD)
           return 0;
        return 1;
      }
  
      /** @return true iff this == x */
      public boolean equals(Object x)
         {return d == ((Element) x).d;}
   }

   /** bounded performance greedy heuristic for 0/1 knapsack problem, k = 2
     * set x[i] = 1 iff object i included in knapsack, 1 <= i < p.length
     * @param p[1:p.length-1] array of profits
     * @param w[1:w.length-1] array of weights
     * @param c knapsack capacity
     * @return value of greedy knapsack filling */
   public static double greedyKnapsack(double [] p, double [] w, double c,
                                       int [] x)
   {
      // define an element array to be sorted by profit density
      Element [] q = new Element [p.length];
      for (int i = 1; i < p.length; i++)
         // array of profit densities
         q[i] = new Element(i, p[i] / w[i]);
   
      // sort by density
      HeapSort.heapSort(q);
   
      int bestfix1 = 0;
      int bestfix2 = 0;
      double maxProfit = 0;
   
      // define object 0 so as to get effect
      // of placing single object subsets into knapsack 
      q[0] = new Element(0, 0.0);
      w[0] = 0;
      p[0] = 0;
   
      // first compute value of best filling
      for (int fix1 = 0; fix1 < p.length - 1; fix1++)
      {//if object q[fix1].id fits into the knapsack put it in
         int fixId1 = q[fix1].id;
         double remainingCapacity1 = c - w[fixId1];
         if (remainingCapacity1 >= 0)
         {// object fixId1 fits
            for (int fix2 = fix1 + 1; fix2 < p.length; fix2++)
            {// try to put object fix2 into the knapsack
               int fixId2 = q[fix2].id;
               double remainingCapacity2 = remainingCapacity1 - w[fixId2];
               if (remainingCapacity2 >= 0)
               {// object fixId2 also fits
                  double currentProfit = p[fixId1] + p[fixId2];
                  // fill remaining capacity in
                  // nonincreasing order of density
                  for (int i = p.length - 1; i > 0; i--)
                  {
                     int id = q[i].id;
                     if (id != fixId1 && id != fixId2
                         && w[id] <= remainingCapacity2)
                     {// object id fits
                        remainingCapacity2 -= w[id];
                        currentProfit += p[id];
                     }
                  }
         
                  if (currentProfit > maxProfit)
                  {// found a better solution
                     maxProfit = currentProfit;
                     bestfix1 = fix1;
                     bestfix2 = fix2;
                  }
               }
            }  
         }     
      }
   
      // now compute the x[] values by redoing the computation
      // with fix1 = bestfix1 and fix2 = bestFix2
      int fixId1 = q[bestfix1].id;
      int fixId2 = q[bestfix2].id;
      double remainingCapacity2 = c - w[fixId1] - w[fixId2];
      x[fixId1] = 1;
      x[fixId2] = 1;
   
      // fill remaining capacity in
      // nonincreasing order of density
      for (int i = p.length - 1; i > 0; i--)
      {
         int id = q[i].id;
         if (id != fixId1 && id != fixId2)
            if (w[id] <= remainingCapacity2)
            {// object id fits
               remainingCapacity2 -= w[id];
               x[id] = 1;
            }
         else
            x[id] = 0;
      }
   
      return maxProfit;
   }
   
   /** test greedyKnapsack */
   public static void main(String [] args)
   {
      double [] p = {0, 4, 3, 10, 12, 8, 10};
      double [] w = {0, 3, 3, 12, 8, 14, 12};
      int [] x = new int [7];
      int n = 6;
      double c = 32;
      System.out.print("Best value is ");
      System.out.println(greedyKnapsack(p, w, c, x));
      System.out.print("x vector is ");
      for (int i = 1; i <= n; i++)
         System.out.print(x[i] + "  ");
      System.out.println();
   }
}
