
/** iterative backtracking solution for knapsack problem
  * computes best loading and its value */

package applications;

public class IterativeBTKnapsack
{
   // top-level nested class
   private static class Element implements Comparable
   {
      // data members
      int id;   // object identifier
      double profitDensity;

      // constructor
      private Element(int theID, double theProfitrofitDensity)
      {
         id = theID;
         profitDensity = theProfitrofitDensity;
      }

      public int compareTo(Object x)
      {
         double xpd = ((Element) x).profitDensity;
         if (profitDensity < xpd) return -1;
         if (profitDensity == xpd) return 0;
         return 1;
      }

      public boolean equals(Object x)
         {return profitDensity == ((Element) x).profitDensity;}
   }

   // class data members
   static double capacity;
   static int numberOfObjects;
   static double [] weight;
   static double [] profit;
   static double weightOfCurrentPacking;
   static double profitFromCurrentPacking;
   static int [] currentPacking;
   static int currentLevel;
   static double weightAtBoundNode;
   static double profitAtBoundNode;
   static int levelAtBoundNode;
   
   /** @param theProfit[1:theProfit.length] is array of object profits
     * @param theWeight[1:theProfit.length] is array of object weights
     * @param theCapacity is knapsack capacity
     * @param thePacking solution array
     * @return profit of best filling */
   public static double knapsack(double [] theProfit, double [] theWeight,
                                 double theCapacity, int [] thePacking)
   {
      numberOfObjects = theProfit.length - 1;
      capacity = theCapacity;

      // define an Element array for profit densities
      Element [] q  = new Element [numberOfObjects];
   
      // set up densities in q[0:numberOfObjects-1]
      for (int i = 1; i <= numberOfObjects; i++)
         q[i - 1] = new Element(i, theProfit[i] / theWeight[i]);
   
      // sort into increasing density order
      MergeSort.mergeSort(q);
   
      // rearrange objects in density order
      profit = new double [numberOfObjects + 1];
      weight = new double [numberOfObjects + 1];
      int [] id = new int [numberOfObjects + 1];  // original id of object i
      for (int i = 1; i <= numberOfObjects; i++)
      {// profits and weights in decreasing density order
         profit[i] = theProfit[q[numberOfObjects - i].id];
         weight[i] = theWeight[q[numberOfObjects - i].id];
         id[i] = q[numberOfObjects - i].id;
      }

      currentPacking = new int [numberOfObjects + 1];
      double maxProfitSoFar = 0.0;
      int saveToIndex = 0;    // incremental saving of
                              // thePacking[1:saveToIndex] yet to be done
      currentLevel = 1;

      // search the tree
      while (true)
      {
         // find a node with a better bound than maxProfitSoFar
         while (bound() <= maxProfitSoFar)
         {
            // backtrack to nearest currentPacking[] = 1
            currentLevel--;   // currentLevel is now nearest
                              // currentPacking[i] that was set
            while (currentLevel != 0 && currentPacking[currentLevel] == 0)
            {// move up one level
               if (saveToIndex == currentLevel)
               {// must save currentPacking[currentLevel]
                  thePacking[id[currentLevel]] = 0;
                  saveToIndex--;
                }
                currentLevel--;
            }   

            if (currentLevel == 0)
            {// we are done
               if (maxProfitSoFar == 0.0)
                  // no object fits
                  for (int i = 1; i <= numberOfObjects; i++)
                     thePacking[i] = 0;
               return maxProfitSoFar;
            }
            
            // move to right child
            if (saveToIndex == currentLevel)
            {// must save currentPacking[currentLevel]
               thePacking[id[currentLevel]] = 1;
               saveToIndex--;
            }
            currentPacking[currentLevel] = 0;
            weightOfCurrentPacking -= weight[currentLevel];
            profitFromCurrentPacking -= profit[currentLevel];
            currentLevel++;
         }
               
         // at a node with a better bound than maxProfitSoFar
         currentLevel = levelAtBoundNode;
         weightOfCurrentPacking = weightAtBoundNode;
         profitFromCurrentPacking = profitAtBoundNode;
         if (currentLevel > numberOfObjects)
         {// at a leaf, need to save all values
            saveToIndex = numberOfObjects;
            maxProfitSoFar = profitFromCurrentPacking;
         }
         else
         {// move to a right child
            currentPacking[currentLevel] = 0;
            currentLevel++;
         }
      }
   }
   
   
   /** bounding function
     * @return upper bound on value of best leaf in subtree
     * rooted at current level currentLevel node,
     * set currentPacking[] = 1 for all left child moves made */
   private static double bound()
   {
      weightAtBoundNode = weightOfCurrentPacking;
            // capacity used at current node
      profitAtBoundNode = profitFromCurrentPacking;
           // profit at current node
      levelAtBoundNode = currentLevel;
           // next object

      // fill remaining capacity in order of profit density
      while (levelAtBoundNode <= numberOfObjects &&
             weightAtBoundNode + weight[levelAtBoundNode] <= capacity)
      {
         weightAtBoundNode += weight[levelAtBoundNode];
         profitAtBoundNode += profit[levelAtBoundNode];
         currentPacking[levelAtBoundNode] = 1;
         levelAtBoundNode++;
      }
   
      // take fraction of next object
      if (levelAtBoundNode <= numberOfObjects)
         return profitAtBoundNode + profit[levelAtBoundNode] /
                  weight[levelAtBoundNode] * (capacity - weightAtBoundNode);
      else
         return profitAtBoundNode;
   }
   
   /** test program */
   public static void main(String [] args)
   {
      double [] p = {0, 6, 3, 5, 4, 6};
      double [] w = {0, 2, 2, 6, 5, 4};
      int n = 5;
      int c = 10;
      int [] bestx = new int [n + 1];
      System.out.println("Optimal value is " + knapsack(p, w, c, bestx));
      System.out.print("The solution vector is ");
      for (int i = 1; i <= n; i++)
         System.out.print(bestx[i] + " " );
      System.out.println();
   }
}
