
/** iterative dynamic programming two-dimensional 0/1 knapsack */

package applications;

import utilities.*;

public class IterativeDP2DKnapsack
{
   /** iterative method to solve dynamic programming recurrence
     * for the two-dimensional 0/1 knapsack problem
     * computes f[1][c][d] and f[i][y][z], 2 <= i <= n, 0 <= y <= c,
     * 0 <= z <= d
     * @param p[1:p.length - 1] gives object profits
     * @param v[1:v.length-1] gives object weights in first dimension
     * @param w[1:w.length-1] gives object weights in second dimension
     * @param c and d are knapsack capacities in each dimension */
   public static void knapsack(int [] p, int [] v, int [] w,
                               int c, int d, int [][][] f)
   {
      int n = p.length - 1;   // number of objects

      // initialize f[n][][]
      // first do y and z such that object n does not fit
      int yMax = Math.min(v[n] - 1, c);
      int zMax = Math.min(w[n] - 1, d);
      for (int y = 0; y <= yMax; y++)
         for (int z = 0; z <= d; z++)
            f[n][y][z] = 0;
      for (int y = v[n]; y <= c; y++)
         for (int z = 0; z <= zMax; z++)
            f[n][y][z] = 0;
   
      // now do y and z such that object n fits
      for (int y = v[n]; y <= c; y++)
         for (int z = w[n]; z <= d; z++)
            f[n][y][z] = p[n];
      
      // compute remaining f's except f[1][][]
      for (int i = n - 1; i > 1; i--)
      {
         // first do for y and z such that
         // object i does not fit
         yMax = Math.min(v[i] - 1, c);
         zMax = Math.min(w[i] - 1, d);
         for (int y = 0; y <= yMax; y++)
            for (int z = 0; z <= d; z++)
               f[i][y][z] = f[i+1][y][z];
         for (int y = v[i]; y <= c ; y++)
            for (int z = 0; z <= zMax; z++)
               f[i][y][z] = f[i + 1][y][z];
   
         // now do for y and z such that object i fits
         for (int y = v[i]; y <= c; y++)
            for (int z = w[i]; z <= d; z++)
               f[i][y][z] = Math.max(f[i + 1][y][z],
                            f[i + 1][y - v[i]][z - w[i]] + p[i]);
      }
   
      // compute f[1][c][d]
      f[1][c][d] = f[2][c][d];
      if (c >= v[1] && d >= w[1])
         // object 1 fits
         f[1][c][d] = Math.max(f[1][c][d],
                      f[2][c - v[1]][d - w[1]] + p[1]);
   
   }
   
   /** compute solution vector x */
   public static void traceback(int [][][] f, int [] v, int [] w,
                                int c, int d, int [] x)
   {
      int n = w.length - 1;   // number of objects
   
      for (int i = 1; i < n; i++)
         if (f[i][c][d] == f[i + 1][c][d])
            x[i] = 0;
         else
         {
            x[i] = 1;
            c -= v[i];
            d -= w[i];
         }
      x[n] = (f[n][c][d] == 0) ? 0 : 1;
   }
   
   /** test program */
   public static void main(String [] args)
   {
      // define the input stream to be the standard input stream
      MyInputStream keyboard = new MyInputStream();

      System.out.println("Enter number of objects and knapsack capacity");
      int n = keyboard.readInteger();
      int c = keyboard.readInteger();
      int d = keyboard.readInteger();
      int [] p = new int [n + 1];
      int [] v = new int [n + 1];
      int [] w = new int [n + 1];
      int [] x = new int [n + 1];
      int [][][] f = new int [n + 1][c + 1][d + 1];

      for (int i = 1; i <= n; i++)
      {
         System.out.println("Enter profit and weight in each"
                            + " dimension of object " + i);
         p[i] = keyboard.readInteger();
         v[i] = keyboard.readInteger();
         w[i] = keyboard.readInteger();
      }

      knapsack(p, v, w, c, d, f);

      System.out.print("Optimal value is ");
      System.out.println(f[1][c][d]);

      traceback(f, v, w, c, d, x);

      System.out.print("The solution vector is ");
      for (int i = 1; i <= n; i++)
         System.out.print(x[i] + " ");
      System.out.println();
   }
}
