
/** dynamic programming knapsack using tuple formulation */

package applications;

import java.util.*;
import dataStructures.*;
import utilities.*;

public class TupleDPKnapsack
{
   // top-level nested class
   private static class Tuple
   {
      // data members
      double weight,       // weight of tuple
             profit;       // profit of tuple

      // methods
      public Tuple(double theWeight, double theProfit)
      {
         weight = theWeight;
         profit = theProfit;
      }

      public String toString()
         {return "(" + weight + ", " + profit + ")";}
   }

   /** iterative method to solve dynamic programming recurrence
     * uses the tuple formulation
     * @param profit[1:profit.length - 1] gives object profits
     * @param weight[1:weight.length - 1] gives object weights
     * @param knapsackCapacity is the knapsack capacity 
     * @param x[1:weight.length - 1] is set to solution vector
     * @return profit from best packing (i.e., of solution vector) */
   public static double knapsack(double [] profit, double [] weight,
                                 double knapsackCapacity, int [] x)
   {
      int numberOfObjects = profit.length - 1;

      // initialize array of tuple lists
      ArrayLinearList [] pairs = new ArrayLinearList [numberOfObjects];
      pairs[0] = new ArrayLinearList();
      pairs[0].add(0, new Tuple(0.0, 0.0));

      // compute pairs[1:numberOfObjects - 1]
      for (int i = 1; i < numberOfObjects; i++)
      {
         // define tuple list for x[i] = 1
         ArrayLinearList q = new ArrayLinearList();

         // create q's tuples from pairs[i - 1]
         Iterator ip = pairs[i - 1].iterator();
         while (ip.hasNext())
         {// process next tuple from pairs[i - 1]
            Tuple t = (Tuple) ip.next();
            if (t.weight + weight[i] > knapsackCapacity)
               // no need to look at remaining tuples of pairs[i - 1]
               // because they have greater weight than current tuple
               break;

            // feasible tuple for q, append
            q.add(q.size(), new Tuple(t.weight + weight[i],
                                      t.profit + profit[i]));
         }

         // initialize pairs[i]
         pairs[i] = new ArrayLinearList();

         // merge q and pairs[i - 1] to get pairs[i]
         // eliminate dominated tuples during the merge
         // note that both lists are in ascending order of both
         // weight and profit
         ip = pairs[i - 1].iterator();
         Iterator iq = q.iterator();
         Tuple pTuple = (Tuple) ip.next();   // pairs[i - 1] has >= 1 pair
         Tuple qTuple = iq.hasNext() ? (Tuple) iq.next() : null;

         while (pTuple != null && qTuple != null)
            // merge tuple with smaller weight into pairs[i]
            // unless this tuple is dominated by another tuple
            if (pTuple.weight < qTuple.weight)
            {// pTuple has smaller weight
               if (pTuple.profit < qTuple.profit)
               {// pTuple is not dominated by any tuple, add to pairs[i]
                  pairs[i].add(pairs[i].size(), pTuple);
                  pTuple = ip.hasNext() ? (Tuple) ip.next() : null;
               }
               else
                  // pTuple dominates qTuple, eliminate qTuple
                  qTuple = iq.hasNext() ? (Tuple) iq.next() : null;
            }
            else
               if (pTuple.weight == qTuple.weight)
               {// one tuple may be eliminated
                  if (pTuple.profit <= qTuple.profit)
                     // pTuple is dominated by qTuple, eliminate pTuple
                     pTuple = ip.hasNext() ? (Tuple) ip.next() : null;
                  else
                     // pTuple dominates qTuple, eliminate qTuple
                     qTuple = iq.hasNext() ? (Tuple) iq.next() : null;
               }
               else
               {// qTuple has smaller weight
                  if (qTuple.profit < pTuple.profit)
                  {// qTuple is not dominated by any tuple, add to pairs[i]
                     pairs[i].add(pairs[i].size(), qTuple);
                     qTuple = iq.hasNext() ? (Tuple) iq.next() : null;
                  }
                  else
                     // qTuple dominates pTuple, eliminate pTuple
                     pTuple = ip.hasNext() ? (Tuple) ip.next() : null;
               }

            // append remaining tuples
            // note that tuples remain in only one list
            while (pTuple != null)
            {
               pairs[i].add(pairs[i].size(), pTuple);
               pTuple = ip.hasNext() ? (Tuple) ip.next() : null;
            }
            while (qTuple != null)
            {
               pairs[i].add(pairs[i].size(), qTuple);
               qTuple = iq.hasNext() ? (Tuple) iq.next() : null;
            }
            
      }                              

      // determine max tuple of pairs[numberOfObjects]
      // first find max tuple if pairs[i - 1] with weight
      // <= knapsackCapacity - weight[numberOfObjects]
      Tuple maxTuple = new Tuple(0.0, 0.0);
      double maxWeight = knapsackCapacity - weight[numberOfObjects];
      // could use binary search instead
      for (int i = pairs[numberOfObjects - 1].size() - 1; i >= 0; i--)
      {
         Tuple nextTuple = (Tuple) pairs[numberOfObjects - 1].get(i);
         if (nextTuple.weight <= maxWeight)
         {// found max tuple
            maxTuple = nextTuple;
            break;
         }
      }

      // determine max profit and x[numberOfObjects]
      double maxProfit = maxTuple.profit + profit[numberOfObjects];
      Tuple t = (Tuple) pairs[numberOfObjects - 1].
                              get(pairs[numberOfObjects - 1].size() - 1);
      if (maxProfit >= t.profit)
         x[numberOfObjects] = 1;
      else
      {// t is best way to go
         x[numberOfObjects] = 0;
         maxTuple = t;
         maxProfit = maxTuple.profit;
      }

      // find remaining x[]s
      for (int i = numberOfObjects - 1; i >= 1; i--)
         // determine x[i]
         if (exists(pairs[i - 1], maxTuple))
            x[i] = 0;
         else
         {
            x[i] = 1;
            maxTuple.weight -= weight[i];
            maxTuple.profit -= profit[i];
         }     

      return maxProfit;
   }

   /** @return true iff x is in tupleList; null otherwise */
   public static boolean exists(LinearList tupleList, Tuple x)
   {// do a binary search for x
      int left = 0;
      int right = tupleList.size() - 1;
      while (left <= right)
      {
         int middle = (left + right)/2;

         double middleWeight = ((Tuple) tupleList.get(middle)).weight;
         if (x.weight > middleWeight)
            left = middle + 1;
         else
            if (x.weight < middleWeight)
               right = middle - 1;
            else
               return true;
      }
      return false; // x not found
   }
   
   /** test program */
   public static void main(String [] args)
   {
      // define the input stream to be the standard input stream
      MyInputStream keyboard = new MyInputStream();

      System.out.println("Enter number of objects and knapsack capacity");
      int n = keyboard.readInteger();
      double c = keyboard.readDouble();
      double [] p = new double [n + 1];
      double [] w = new double [n + 1];
      int [] x = new int [n + 1];

      for (int i = 1; i <= n; i++)
      {
         System.out.println("Enter profit and weight of object " + i);
         p[i] = keyboard.readDouble();
         w[i] = keyboard.readDouble();
      }

      System.out.println("Optimal value is " + knapsack(p, w, c, x));
      System.out.print("Solution vector is ");
      System.out.println();
      for (int i = 1; i <= n; i++)
         System.out.print(x[i] + " ");
      System.out.println();
   }
}
