
/** backtracking method to find a max clique
  * start by sorting vertices by degree */

package dataStructures;

import applications.*;
import utilities.*;
import wrappers.*;

public class AdjacencyGraphWithSortedMaxClique extends AdjacencyGraph
{
   // top-level nested class
   private static class Element implements Comparable
   {
      // data members
      int vertex; 
      int degree;        // vertex degree

      // constructor
      private Element(int theVertex, int theDegree)
      {
         vertex = theVertex;
         degree = theDegree;
      }

      // method of Comparable, inverts normal order
      public int compareTo(Object x)
      {
         int xDegree = ((Element) x).degree;
         if (degree > xDegree)
            return -1;
         if (degree == xDegree)
            return 0;
         return 1;
      }

      /** @return true iff this = x */
      public boolean equals(Object x)
         {return degree == ((Element) x).degree;}
   }

   // constructors
   public AdjacencyGraphWithSortedMaxClique(int theVertices)
      {super(theVertices);}

   public AdjacencyGraphWithSortedMaxClique()
      {this(0);}

   // class data members used by backtracking max clique
   static int [] currentClique;
   static int sizeOfCurrentClique;
   static int sizeOfMaxCliqueSoFar;
   static int [] maxCliqueSoFar;
   static Element [] sortedVertices;  // vertices sorted by degree
   
   /** solve max-clique problem using backtracking
     * @return size of max clique
     * @param maxClique set maxClique[i] = 1 iff i is in max clique */
   public int btMaxClique(int [] maxClique)
   {
      // define an Element array for vertex degrees
      sortedVertices  = new Element [n + 1];
   
      // set up degrees in sortedVertices[1:n]
      for (int i = 1; i <= n; i++)
         sortedVertices[i] = new Element(i, degree(i));
   
      // sort into increasing density order
      HeapSort.heapSort(sortedVertices);

      // initialize for rClique
      currentClique = new int [n + 1];
      sizeOfCurrentClique = 0;
      sizeOfMaxCliqueSoFar = 0;
      maxCliqueSoFar = maxClique;
   
      // find max clique
      rClique(1);
      return sizeOfMaxCliqueSoFar;
   }
   
   /** recursive backtracking code to compute largest clique */
   private void rClique(int currentLevel)
   {// search from a node at currentLevel
      if (currentLevel > n)
      {// at leaf, found a larger clique
       // update maxCliqueSoFar and sizeOfMaxCliqueSoFar
         for (int j = 1; j <= n; j++)
            maxCliqueSoFar[sortedVertices[j].vertex] = currentClique[j];
         sizeOfMaxCliqueSoFar = sizeOfCurrentClique;
         return;
      }
   
      // not at leaf, see if vertex currentLevel
      // is connected to others in current clique
      boolean connected = true;
      for (int j = 1; j < currentLevel; j++)
         if (currentClique[j] == 1 &&
             !a[sortedVertices[currentLevel].vertex][sortedVertices[j].vertex])
         {// vertex currentLevel not connected to j
            connected = false;
            break;
         }
   
      if (connected)
      {// try left subtree
         currentClique[currentLevel] = 1;  // add to clique
         sizeOfCurrentClique++;
         rClique(currentLevel + 1);
         sizeOfCurrentClique--;
      }
   
      if (sizeOfCurrentClique + n - currentLevel > sizeOfMaxCliqueSoFar)
      {// try right subtree
         currentClique[currentLevel] = 0;
         rClique(currentLevel + 1);
      }
   }

   public static void main(String [] args)
   {
      int n = 7;
      AdjacencyGraph g = new AdjacencyGraph(n);

      // define a MyInputStream object to input from System.in
      MyInputStream keyboard = new MyInputStream();

      // input a test graph
      System.out.println("enter number of edges in 7 vertex graph");
      int e = keyboard.readInteger();
      for (int i = 1; i <= e; i++)
      {
         System.out.println("enter unweighted edge " + i);
         int u = keyboard.readInteger();
         int v = keyboard.readInteger();
         g.putEdge(new Edge(u, v));
      }

      System.out.println("The undirected graph is");
      g.output();

      // test btMaxClique
      int [] v = new int [n + 1];
      System.out.println("\nMax Clique size is " + g.btMaxClique(v));
      System.out.print("\nSolution vector is ");
      for (int i = 1; i <= n; i++)
         System.out.print(v[i] + " ");
      System.out.println();
   }
}
