

/** hash tables using array linear lists and division, from scratch */

package dataStructures;

import java.util.*;
import utilities.*;

public class ArrayHashTable
{
   // top-level nested class
   protected static class HashEntry
   {
      // data members
      protected Comparable key;
      protected Object element;

      // constructors
      private HashEntry() {}
     
      private HashEntry(Comparable theKey, Object theElement)
      {
         key = theKey;
         element = theElement;
      }
   }

   // data members
   private int divisor;                     // hash function divisor
   private HashEntry [][] table;            // hash table array
   private int [] bucketSize;               // number of elements in a bucket
   private int size;                        // number of elements in table

   // constructor
   public ArrayHashTable(int theDivisor)
   {
      divisor = theDivisor;
   
      // allocate arrays
      table = new HashEntry [divisor][10];
      bucketSize = new int [divisor];
   }
   
   // instance methods
   /** @return true iff the hash table is empty */
   public boolean isEmpty()
       {return size == 0;}

   /** @return current number of elements in the hash table */
   public int size()
      {return size;}

   /** @return element with specified key
     * @return null if no matching element */
   public Object get(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket
   
      // search from left to right for matching element
      // bucket size is expected to be small
      // so binary search is not used
      int i = 0, bSize = bucketSize[b];
      HashEntry [] bucket = table[b];
      while (i < bSize && bucket[i].key.compareTo(theKey) < 0)
         i++;
   
      // verify match
      if (i >= bSize || !bucket[i].key.equals(theKey))
         // no matching element
         return null;
      else
         // there is a match
         return bucket[i].element;
   }
   
   /** insert an element with the specified key
     * overwrite old element if there is already an
     * element with the given key 
     * @return old element (if any) with key = theKey*/
   public Object put(Object theKey, Object theElement)
   {
      int b = theKey.hashCode() % divisor;  // home bucket

      // search from left to right for matching element
      // bucket size is expected to be small
      // so binary search is not used
      int i = 0, bSize = bucketSize[b];
      HashEntry [] bucket = table[b];
      while (i < bSize && bucket[i].key.compareTo(theKey) < 0)
         i++;
   
      // verify match
      if (i < bSize && bucket[i].key.equals(theKey))
      {// replace old element
         Object elementToReturn = bucket[i].element;
         bucket[i].element = theElement;
         return elementToReturn;
      }

      if (bSize == bucket.length)
      {// no space for new element, double capacity
         bucket = (HashEntry []) ChangeArrayLength.changeLength1D
                      (bucket, 2 * bucket.length);
         table[b] = bucket;
      }
   
      // create space for new element by moving
      // [i:bSize - 1] one spot right
      System.arraycopy(bucket, i, bucket, i + 1, bSize - i);

      // insert at bucket[i]
      bucket[i] = new HashEntry((Comparable) theKey, theElement);
      bucketSize[b]++;
      size++;
      return null;
   }

   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket

      // search for match with theKey
      // bucket size is expected to be small
      // so binary search is not used
      int i = 0, bSize = bucketSize[b];
      HashEntry [] bucket = table[b];
      while (i < bSize && bucket[i].key.compareTo(theKey) < 0)
         i++;
   
      // verify match
      if (i >= bSize || !bucket[i].key.equals(theKey))
         // no matching element
         return null;
   
      // save matching element and remove from list
      Object removedElement = bucket[i].element;
      System.arraycopy(bucket, i + 1, bucket, i, bSize - i - 1);
      bucketSize[b]--;  // new bucket size
      size--;           // new table size

      return removedElement;
   }
   
   /** output the hash table */
   public void output()
   {
      for (int i = 0; i < divisor; i++)
      {
         // output the array table[i]
         HashEntry [] bucket = table[i];
         System.out.print("[");
         for (int j = 0; j < bucketSize[i]; j++)
         {
            if (j != 0)
               // not first element
               System.out.print(", ");
            System.out.print(bucket[j].element.toString());
         }
         System.out.println("]");
     }

      System.out.println("Table size is " + size);
   }
   
   /** test method */
   public static void main (String [] args)
   {
      ArrayHashTable h = new ArrayHashTable(11);
      h.put(new Integer(80), new Integer(80));
      h.put(new Integer(40), new Integer(40));
      h.put(new Integer(65), new Integer(65));
      h.output();
      System.out.println();
      h.put(new Integer(58), new Integer(58));
      h.put(new Integer(24), new Integer(24));
      h.output();
      System.out.println();
      h.put(new Integer(2), new Integer(2));
      h.put(new Integer(13), new Integer(13));
      h.put(new Integer(46), new Integer(46));
      h.put(new Integer(16), new Integer(16));
      h.put(new Integer(7), new Integer(7));
      h.put(new Integer(21), new Integer(21));
      h.output();
      System.out.println();
      h.put(new Integer(99), new Integer(99));

      // update element
      h.put(new Integer(7), new Integer(29));
      h.output();
   }
}
