/** final version of hash tables using array linear lists and division
  * includes array doubling, from scratch */

package dataStructures;

import java.util.*;
import utilities.*;

public class FinalArrayHashTable
{
   // top-level nested class
   protected static class HashEntry
   {
      // data members
      protected Comparable key;
      protected Object element;

      // constructors
      private HashEntry() {}
     
      private HashEntry(Comparable theKey, Object theElement)
      {
         key = theKey;
         element = theElement;
      }
   }

   // data members
   private int divisor;                     // hash function divisor
   private HashEntry [][] table;            // hash table array
   private int [] bucketSize;               // number of elements in a bucket
   private int size;                        // number of elements in table
   private double threshold;
         // double hash table capacity when loading factor
         // exceeds threshold

   // constructor
   public FinalArrayHashTable(int theDivisor, double theThreshold)
   {
      if (theDivisor < 1)
         throw new IllegalArgumentException("Divisor must be >= 1.");

      divisor = theDivisor;
      threshold = theThreshold;
   
      // allocate arrays
      table = new HashEntry [divisor][1];
      bucketSize = new int [divisor];
   }

   public FinalArrayHashTable(double theThreshold)
      {this(11, theThreshold);}

   public FinalArrayHashTable(int theDivisor)
      {this(theDivisor, 0.75);}

   public FinalArrayHashTable()
      {this(11, 0.75);}

   // instance methods
   /** @return true iff the hash table is empty */
   public boolean isEmpty()
       {return size == 0;}

   /** @return current number of elements in the hash table */
   public int size()
      {return size;}

   /** @return element with specified key
     * @return null if no matching element */
   public Object get(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket
   
      // search from left to right for matching element
      // bucket size is expected to be small
      // so binary search is not used
      int i = 0, bSize = bucketSize[b];
      HashEntry [] bucket = table[b];
      while (i < bSize && bucket[i].key.compareTo(theKey) < 0)
         i++;
   
      // verify match
      if (i >= bSize || !bucket[i].key.equals(theKey))
         // no matching element
         return null;
      else
         // there is a match
         return bucket[i].element;
   }
   
   /** insert an element with the specified key
     * overwrite old element if there is already an
     * element with the given key 
     * @return old element (if any) with key = theKey*/
   public Object put(Object theKey, Object theElement)
   {
      if (((double)(size + 1)) / table.length > threshold)
      {// double table capacity, does an unnecessary doubling when theKey
       // is a duplicate, doesn't hurt
         HashEntry [][] oldTable = table;
         int [] oldBucketSize = bucketSize;
         // approximately double capacity, keeping length odd 
         table = new HashEntry [oldTable.length * 2 + 1][1];
         divisor = table.length;
         bucketSize = new int [divisor];

         // insert old elements into new table
         for (int i = 0; i < oldTable.length; i++)
         {
            int obSize = oldBucketSize[i];
            for (int j = 0; j < obSize; j++)
            {// enter oldTable[i][j] into new table
               // find home bucket for old element
               Comparable oldKey = oldTable[i][j].key;
               int b = Math.abs(oldKey.hashCode()) % divisor;

               // find place in new home bucket
               int k = 0, bSize = bucketSize[b];
               HashEntry [] bucket = table[b];
               while (k < bSize && bucket[k].key.compareTo(oldKey) < 0)
                  k++;
            
               if (bSize == bucket.length)
               {// no space for old element, double capacity
                  bucket = (HashEntry []) ChangeArrayLength.changeLength1D
                               (bucket, 2 * bucket.length);
                  table[b] = bucket;
               }
            
               // create space for new element by moving
               // [k:bSize - 1] one spot right
               System.arraycopy(bucket, k, bucket, k + 1, bSize - k);
         
               // insert at bucket[k]
               bucket[k] = oldTable[i][j];
               bucketSize[b]++;
            }
         }
      }

      // insert new element
      int b = theKey.hashCode() % divisor;  // home bucket

      // search from left to right for matching element
      // bucket size is expected to be small
      // so binary search is not used
      int i = 0, bSize = bucketSize[b];
      HashEntry [] bucket = table[b];
      while (i < bSize && bucket[i].key.compareTo(theKey) < 0)
         i++;
   
      // verify match
      if (i < bSize && bucket[i].key.equals(theKey))
      {// replace old element
         Object elementToReturn = bucket[i].element;
         bucket[i].element = theElement;
         return elementToReturn;
      }

      if (bSize == bucket.length)
      {// no space for new element, double capacity
         bucket = (HashEntry []) ChangeArrayLength.changeLength1D
                      (bucket, 2 * bucket.length);
         table[b] = bucket;
      }
   
      // create space for new element by moving
      // [i:bSize - 1] one spot right
      System.arraycopy(bucket, i, bucket, i + 1, bSize - i);

      // insert at bucket[i]
      bucket[i] = new HashEntry((Comparable) theKey, theElement);
      bucketSize[b]++;
      size++;
      return null;
   }

   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket

      // search for match with theKey
      // bucket size is expected to be small
      // so binary search is not used
      int i = 0, bSize = bucketSize[b];
      HashEntry [] bucket = table[b];
      while (i < bSize && bucket[i].key.compareTo(theKey) < 0)
         i++;
   
      // verify match
      if (i >= bSize || !bucket[i].key.equals(theKey))
         // no matching element
         return null;
   
      // save matching element and remove from list
      Object removedElement = bucket[i].element;
      System.arraycopy(bucket, i + 1, bucket, i, bSize - i - 1);
      bucketSize[b]--;  // new bucket size
      size--;           // new table size

      return removedElement;
   }
   
   /** output the hash table */
   public void output()
   {
      for (int i = 0; i < divisor; i++)
      {
         // output the array table[i]
         HashEntry [] bucket = table[i];
         System.out.print("[");
         for (int j = 0; j < bucketSize[i]; j++)
         {
            if (j != 0)
               // not first element
               System.out.print(", ");
            System.out.print(bucket[j].element.toString());
         }
         System.out.println("]");
     }

      System.out.println("Table size is " + size);
   }
   
   /** test method */
   public static void main (String [] args)
   {
      FinalArrayHashTable h = new FinalArrayHashTable();
      h.put(new Integer(80), new Integer(80));
      h.put(new Integer(40), new Integer(40));
      h.put(new Integer(65), new Integer(65));
      h.put(new Integer(58), new Integer(58));
      h.put(new Integer(24), new Integer(24));
      h.put(new Integer(2), new Integer(2));
      h.put(new Integer(13), new Integer(13));
      h.put(new Integer(46), new Integer(46));
      h.put(new Integer(16), new Integer(16));
      h.put(new Integer(7), new Integer(7));
      h.put(new Integer(21), new Integer(21));
      h.put(new Integer(99), new Integer(99));
      h.put(new Integer(17), new Integer(17));
      h.put(new Integer(35), new Integer(35));
      h.put(new Integer(54), new Integer(54));
      h.output();
      System.out.println();
      h.remove(new Integer(7));
      h.remove(new Integer(40));
      h.remove(new Integer(16));
      h.output();
   }
}
