
/** final version of hash tables using chaining and division
  * includes doubling
  * developed from basics */

package dataStructures;

public class FinalChainedHashTable
{
   // top-level nested class
   static class HashNode
   {
      // data members
      Comparable key;
      Object element;
      HashNode next;

      // constructors
      HashNode() {}
     
      HashNode(Comparable theKey, Object theElement)
      {
         key = theKey;
         element = theElement;
      }
   }

   // data members of FinalChainedHashTable
   int divisor;               // hash function divisor
   HashNode [] table;         // hash table array
   int size;                  // number of elements in table
   double threshold;
         // double hash table capacity when loading factor
         // exceeds threshold

   // constructor
   public FinalChainedHashTable(int theDivisor, double theThreshold)
   {
      if (theDivisor < 1)
         throw new IllegalArgumentException("Divisor must be >= 1.");

      divisor = theDivisor;
      threshold = theThreshold;
   
      // allocate hash table array
      table = new HashNode [divisor];
   }

   public FinalChainedHashTable(double theThreshold)
      {this(11, theThreshold);}

   public FinalChainedHashTable(int theDivisor)
      {this(theDivisor, 0.75);}

   public FinalChainedHashTable()
      {this(11, 0.75);}
   
   // instance methods
   /** @return true iff the hash table is empty */
   public boolean isEmpty()
      {return size == 0;}

   /** @return current number of elements in the hash table */
   public int size()
      {return size;}

   /** @return element with specified key
     * @return null if no matching element */
   public Object get(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket
   
      // search home bucket chain
      HashNode currentNode = table[b];
      while (currentNode != null && currentNode.key.compareTo(theKey) < 0)
         currentNode = currentNode.next;

      // did we find a matching element?
      if (currentNode == null || !currentNode.key.equals(theKey))
         // no
         return null;
      else
         // yes
         return currentNode.element;
   }
   
   /** insert an element with the specified key
     * overwrite old element if there is already an
     * element with the given key 
     * double table size when loading factor exceeds threshold
     * @return old element (if any) with key = theKey*/
   public Object put(Object theKey, Object theElement)
   {
      if (((double)(size + 1)) / table.length > threshold)
      {// double table capacity, does an unnecessary doubling when theKey
       // is a duplicate, doesn't hurt
         HashNode [] oldTable = table;
         // approximately double capacity, keeping length odd 
         table = new HashNode [oldTable.length * 2 + 1];
         divisor = table.length;

         // insert old elements into new table
         for (int i = 0; i < oldTable.length; i++)
         {
            HashNode oldNode = oldTable[i];
            while (oldNode != null)
            {// enter oldNode into new table
               // find home bucket for oldNode
               int b = Math.abs(oldNode.key.hashCode()) % divisor;
      
               // search home bucket chain of new table
               HashNode currentNode = table[b],
                        previousNode = null;
         
               while (currentNode != null &&
                      currentNode.key.compareTo(theKey) < 0)
               {// move to next node on home bucket chain
                  previousNode = currentNode;
                  currentNode = currentNode.next;
               }
            
               // insert oldNode after previousNode and before currentNode
               HashNode newNode = oldNode;
               oldNode = oldNode.next;
               newNode.next = currentNode;
               if (previousNode == null)
                  // insert at front
                  table[b] = newNode;
               else
                  // not at front
                  previousNode.next = newNode;
            }
         }
      }

      // insert new element
      int b = theKey.hashCode() % divisor;  // home bucket

      // search home bucket chain
      HashNode currentNode = table[b],
               previousNode = null;

      while (currentNode != null && currentNode.key.compareTo(theKey) < 0)
      {// move to next node on home bucket chain
         previousNode = currentNode;
         currentNode = currentNode.next;
      }
   
      // check if element with theKey already in table
      if (currentNode != null && currentNode.key.equals(theKey))
      {// update the element
         Object elementToReturn = currentNode.element;
         currentNode.element = theElement;
         return elementToReturn;
      }
   
      // insert theElement after previousNode and before currentNode
      HashNode newNode = new HashNode((Comparable) theKey, theElement);
      newNode.next = currentNode;
      if (previousNode == null)
         // insert at front
         table[b] = newNode;
      else
         // not at front
         previousNode.next = newNode;

      size++;
      return null;
   }

   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket

      // search the home bucket for matching element
      // using a trailing pointer previousNode
      HashNode currentNode = table[b],
               previousNode = null;
      while (currentNode != null && currentNode.key.compareTo(theKey) < 0)
      {// move to next node on home bucket chain
         previousNode = currentNode;
         currentNode = currentNode.next;
      }
   
      // check if we have a match
      if (currentNode == null || !currentNode.key.equals(theKey))
         // no element with theKey in table
         return null;
   
      // we have a match, save element to be removed
      Object elementToReturn = currentNode.element;
      if (previousNode != null)
         // not removing first node from home bucket chain
         previousNode.next = currentNode.next;
      else
         // remove first node from home bucket chain
         table[b] = currentNode.next;

      size--;
      return elementToReturn;
   }
   
   /** output the hash table */
   public void output()
   {
      for (int i = 0; i < divisor; i++)
      {
         // output the chain table[i]
         HashNode currentNode = table[i];
         System.out.print("[");
         while(currentNode != null)
         {
            if (currentNode != table[i])
               // not first element
               System.out.print(", ");
            System.out.print(currentNode.element.toString());
            currentNode = currentNode.next;
         }
         System.out.println("]");
     }

      System.out.println("Table size is " + size);
   }
   
   /** test method */
   public static void main (String [] args)
   {
      FinalChainedHashTable h = new FinalChainedHashTable();
      h.put(new Integer(80), new Integer(80));
      h.put(new Integer(40), new Integer(40));
      h.put(new Integer(65), new Integer(65));
      h.put(new Integer(58), new Integer(58));
      h.put(new Integer(24), new Integer(24));
      h.put(new Integer(2), new Integer(2));
      h.put(new Integer(13), new Integer(13));
      h.put(new Integer(46), new Integer(46));
      h.put(new Integer(16), new Integer(16));
      h.put(new Integer(7), new Integer(7));
      h.put(new Integer(21), new Integer(21));
      h.put(new Integer(99), new Integer(99));
      h.put(new Integer(17), new Integer(17));
      h.put(new Integer(35), new Integer(35));
      h.put(new Integer(54), new Integer(54));
      h.output();
      System.out.println();
      h.remove(new Integer(7));
      h.remove(new Integer(40));
      h.remove(new Integer(16));
      h.output();
   }
}
