
/** hash tables using linear open addressing and division
  * glorified version includes doubling and remove */

package dataStructures;

import java.util.*;

public class FinalHashTable
{
   // top-level nested class
   protected static class HashEntry
   {
      // data members
      protected Object key;
      protected Object element;

      // constructors
      private HashEntry() {}
     
      private HashEntry(Object theKey, Object theElement)
      {
         key = theKey;
         element = theElement;
      }
   }

   // data members of FinalHashTable
   protected int divisor;         // hash function divisor
   protected HashEntry [] table;  // hash table array
   protected int size;            // number of elements in table
   protected double threshold;
         // double hash table capacity when loading factor
         // exceeds threshold

   // constructor
   public FinalHashTable(int theDivisor, double theThreshold)
   {
      if (theDivisor < 1)
         throw new IllegalArgumentException("Divisor must be >= 1.");

      if (theThreshold > 1.0)
         throw new IllegalArgumentException("Threshold must be <= 1.");

      divisor = theDivisor;
      threshold = theThreshold;
   
      // allocate hash table array
      table = new HashEntry [divisor];
   }

   public FinalHashTable(double theThreshold)
      {this(11, theThreshold);}

   public FinalHashTable(int theDivisor)
      {this(theDivisor, 0.75);}

   public FinalHashTable()
      {this(11, 0.75);}

   // methods
   /** @return true iff the table is empty */
   public boolean isEmpty()
      {return size == 0;}

   /** @return current number of elements in the table */
   public int size()
      {return size;}
   
   /** search an open addressed hash table for an element with key theKey
     * @return location of matching element if found, otherwise return
     * location where an element with key theKey may be inserted
     * provided the hash table is not full */
   private int search(Object theKey)
   {
      int i = Math.abs(theKey.hashCode()) % divisor;  // home bucket
      int j = i;    // start at home bucket
      
      do
      {
         if (table[j] == null || table[j].key.equals(theKey))
            return j;
         j = (j + 1) % divisor;  // next bucket
      } while (j != i);          // returned to home bucket?
   
      return j;  // table full
   }
   
   /** @return element with specified key
     * @return null if no matching element */
   public Object get(Object theKey)
   {
      // search the table
      int b = search(theKey);

      // see if a match was found at table[b]
      if (table[b] == null || !table[b].key.equals(theKey))
         return null;           // no match

      return table[b].element;  // matching element
   }
   
   /** insert an element with the specified key
     * overwrite old element if there is already an
     * element with the given key
     * double table size when loading factor exceeds threshold
     * @return old element (if any) with key theKey */
   public Object put(Object theKey, Object theElement)
   {
      if (((double)(size + 1)) / table.length > threshold)
      {// double table capacity, does an unnecessary doubling when theKey
       // is a duplicate, doesn't hurt
         HashEntry [] oldTable = table;
         // approximately double capacity, keeping length odd 
         table = new HashEntry [oldTable.length * 2 + 1];
         divisor = table.length;

         // insert old elements into new table
         for (int i = 0; i < oldTable.length; i++)
            if (oldTable[i] != null)
            {// enter oldTable[i] into new table
                // find an empty bucket
                int j = Math.abs(oldTable[i].key.hashCode()) % divisor;
                while (table[j] != null)
                   j = (j + 1) % divisor;  // next bucket
      
                table[j] = oldTable[i];
            }
      }

      // search the table for a matching element
      int b = search(theKey);
   
      // check if matching element found
      if (table[b] == null)
      {
         // no matching element
         table[b] = new HashEntry(theKey, theElement);
         size++;
         return null;
      }
      else
      {// must be duplicate
         // change table[b].element
         Object elementToReturn = table[b].element;
         table[b].element = theElement;
         return elementToReturn;
      }
   }
   
   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Object theKey)
   {
      // search for matching element
      int b = search(theKey);

      // check if we have a match
      if (table[b] == null || !table[b].key.equals(theKey))
         // no element with theKey in table
         return null;
   
      // we have a match, save element to be removed
      Object elementToReturn = table[b].element;
      table[b] = null;

      // move elements as needed
      int nextBucket = (b + 1) % divisor;
      while (table[nextBucket] != null)
      {
         int home = Math.abs(table[nextBucket].key.hashCode()) % divisor;
               // home bucket
         if (!(
               (b < home && home <= nextBucket) ||
               (nextBucket < b && b < home) ||
               (home <= nextBucket && nextBucket < b)))
         {// move element in table[nextBucket]
            table[b] = table[nextBucket];
            table[nextBucket] = null;
            b = nextBucket;  // empty bucket
         }
         nextBucket = (nextBucket + 1) % divisor;
      }
      size--;
      return elementToReturn;
   }
   
   /** output the hash table */
   public void output()
   {
      for (int i = 0; i < divisor; i++)
         if (table[i] == null)
            System.out.println("null");
         else
            System.out.println(table[i].element);
   }

   /** test method */
   public static void main (String [] args)
   {
      FinalHashTable h = new FinalHashTable();
      h.put(new Integer(80), new Integer(80));
      h.put(new Integer(40), new Integer(40));
      h.put(new Integer(65), new Integer(65));
      h.put(new Integer(58), new Integer(58));
      h.put(new Integer(24), new Integer(24));
      h.put(new Integer(2), new Integer(2));
      h.put(new Integer(13), new Integer(13));
      h.put(new Integer(46), new Integer(46));
      h.put(new Integer(16), new Integer(16));
      h.put(new Integer(7), new Integer(7));
      h.put(new Integer(21), new Integer(21));
      h.put(new Integer(99), new Integer(99));
      h.put(new Integer(17), new Integer(17));
      h.put(new Integer(35), new Integer(35));
      h.put(new Integer(54), new Integer(54));
      h.output();
      System.out.println();
      h.remove(new Integer(7));
      h.remove(new Integer(40));
      h.remove(new Integer(16));
      h.output();
   }
}
