/** array-based loser trees using full binary trees,
  * initialization done as per Exercise 14.17 (c) */

package dataStructures;

import utilities.*;
import exceptions.*;

public class FullLoserTree
{
   // data members
   int offset;           // 2^ceil(log n) - 1
   int [] tree;          // array for loser tree
   Playable [] player;   // array of players

   // only default constructor available

   /** @return the winner of the tournament
     * @return 0 if there are no players */
   public int getWinner()
      {return (tree == null) ? 0 : tree[0];}

   /** initialize loser tree for thePlayer[1:thePlayer.length-1] */
   public void initialize(Playable [] thePlayer)
   {
      int n = thePlayer.length - 1;
      if (n < 2)
         throw new IllegalArgumentException
                   ("must have at least 2 players");

      // compute  smallest m >= n and a power of 2
      int m;
      for (m = 1; m < n; m += m);
   
      // set instance data members
      player = thePlayer;
      tree = new int [m];
      offset = m - 1;
   
      // play all matches other than those on the rightmost path
      for (int i = 2; i < n; i += 2)
         play((offset + i) / 2, i - 1, i);
   
      // play matches on rightmost path
      int p = (offset + n) / 2;
      if (n % 2 == 1)
         // n is odd, no match at tree[p]
         playToRoot(p / 2, n);
      else 
         // n is even
         if (player[n - 1].winnerOf(player[n]))
         {// n - 1 is winner
            tree[p] = n;
            playToRoot(p / 2, n - 1);
         }
         else
         {// n is winner
            tree[p] = n - 1;
            playToRoot(p / 2, n);
         }
   }

   /** play matches beginning at tree[p]
     * @param lc is left child of p
     * @param rc is right child of p */
   void play(int p, int lc, int rc)
   {
      int currentWinner = player[lc].winnerOf(player[rc]) ? lc : rc;
      if (currentWinner == lc)
         // loser is rc
         tree[p] = rc;
      else // loser is lc
         tree[p] = lc;
   
      // more matches possible if at right child
      while (p > 1 && p % 2 == 1)
      {// at a right child
         // parent of p has opponent information
         p /= 2;  // go to parent
         int newWinner = player[tree[p]].winnerOf(player[currentWinner]) ?
                                tree[p] : currentWinner;
         if (newWinner != currentWinner)
         {// loser is currentWinner
            tree[p] = currentWinner;
            currentWinner = newWinner;
         }
      }
   
      // record winner of last match in tree[p/2]
      tree[p / 2] = currentWinner;
   }

   /** play matches on rightmost path beginning at tree[p / 2]
     * @param currentWinner is one of the players */
   void playToRoot(int p, int currentWinner)
   {
      while (p > 1)
      {
         if (p % 2 == 1)
         {// at a right child
            // parent of p has opponent information
            p /= 2;  // go to parent
            int newWinner = player[tree[p]].winnerOf(player[currentWinner]) ?
                                tree[p] : currentWinner;
            if (newWinner != currentWinner)
            {// loser is currentWinner
               tree[p] = currentWinner;
               currentWinner = newWinner;
            }
         }
         else
            // at a left child, no match to play
            p /= 2;
      }
   
      // record over all winner in tree[0]
      tree[0] = currentWinner;
   }
   
   /** replay matches for previous winner */
   public void rePlay()
   {
      if (player == null)
         throw new IllegalArgumentException
                   ("tree must be initialized first");

      // play matches to root
      int currentWinner = tree[0];
      int p = (currentWinner + offset) / 2;  // node for first match
      while (p >= 1)
      {
         if (tree[p] != 0)
            // two players, play the match
            if (!player[currentWinner].winnerOf(player[tree[p]]))
            {// currentWinner loses the match at tree[p]
               int temp = tree[p];
               tree[p] = currentWinner;
               currentWinner = temp;
            }
             
         p /= 2;  // move to parent
      }

      tree[0] = currentWinner;
   }
   
   public void output()
   {
      int n = player.length - 1;
      System.out.println("size = " + n  + " offset = "  + offset);
      System.out.println("Loser tree pointers are");
      for (int i = 0; i <= offset; i++)
         System.out.print(tree[i] + " ");
      System.out.println();
   }

   /** test program */
   public static void main(String [] args)
   {
      // define a MyInputStream object to input from System.in
      MyInputStream keyboard = new MyInputStream();
   
      // use two initial data sets
      for (int j = 1; j <= 2; j++)
      {
         // get data from System.in
         System.out.println("Enter number of players");
         int n = keyboard.readInteger();
         if (n < 2)
            throw new MyInputException
                      ("must have >= 2 players");
   
         Player [] thePlayer = new Player [n + 1];
   
         System.out.println("Enter player values");
         for (int i = 1; i <= n; i++)
            thePlayer[i] = new Player(i, keyboard.readInteger());
   
         FullLoserTree w = new FullLoserTree();
         w.initialize(thePlayer);
   
         System.out.println("The loser tree is");
         w.output();
   
         // change player n's value
         thePlayer[n].value = 20;
         w.rePlay();
         System.out.println("Changed player " + n + " to 20, new tree is");
         w.output();
   
         // change player 3's value
         thePlayer[3].value = 7;
         w.rePlay();
         System.out.println("Changed player 3 to 7, new tree is");
         w.output();
    
         // change player 1's value
         thePlayer[1].value = 25;
         w.rePlay();
         System.out.println("Changed player 1 to 25, new tree is");
         w.output();
      }
   }
}
