

/** hash tables using chaining and division, developed from basics
  * using chains with a common tail node */

package dataStructures;

import java.util.*;
import wrappers.*;
import utilities.*;

public class HashChainWithTail
{
   // top-level nested class
   private static class HashNode
   {
      // data members
      Comparable key;
      Object element;
      HashNode next;

      // constructors
      private HashNode() {}
     
      private HashNode(Comparable theKey, Object theElement)
      {
         key = theKey;
         element = theElement;
      }
   }

   // data members of HashChainWithTail
   private int divisor;               // hash function divisor
   private HashNode [] table;         // hash table array
   private HashNode tailNode;         // common tail node
   private int size;                  // number of elements in table

   // constructor
   public HashChainWithTail(int theDivisor)
   {
      divisor = theDivisor;
      tailNode = new HashNode();
      // size has default initial value 0
   
      // allocate hash table array
      table = new HashNode [divisor];

      // every chain has a tail node on it
      for (int i = 0; i < divisor; i++)
         table[i] = tailNode;
   }
   
   // instance methods
   /** @return true iff the hash table is empty */
   public boolean isEmpty()
      {return size == 0;}

   /** @return current number of elements in the hash table */
   public int size()
      {return size;}

   /** @return element with specified key
     * @return null f no matching element */
   public Object get(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket
   
      // search home bucket chain
      // first put theKey into the tailnode
      tailNode.key = (Comparable) theKey;
      HashNode currentNode = table[b];
      while (currentNode.key.lessThan(theKey))
         currentNode = currentNode.next;

      // did we find a matching element?
      if (!currentNode.key.equals(theKey) || currentNode == tailNode)
         // no
         return null;
      else
         // yes
         return currentNode.element;
   }
   
   /** insert an element with the specified key
     * overwrite old element if there is already an
     * element with the given key */
   public void put(Object theKey, Object theElement)
   {
      int b = theKey.hashCode() % divisor;  // home bucket

      // search home bucket chain
      // first put theKey into the tailnode
      tailNode.key = (Comparable) theKey;
      HashNode currentNode = table[b],
               previousNode = null;

      while (currentNode.key.lessThan(theKey))
      {// move to next node on home bucket chain
         previousNode = currentNode;
         currentNode = currentNode.next;
      }
   
      // check if element with theKey already in table
      if (currentNode != tailNode && currentNode.key.equals(theKey))
      {// update the element
         currentNode.element = theElement;
         return;
      }
   
      // insert theElement after previousNode and before currentNode
      HashNode newNode = new HashNode((Comparable) theKey, theElement);
      newNode.next = currentNode;
      if (previousNode == null)
         // insert at front
         table[b] = newNode;
      else
         // not at front
         previousNode.next = newNode;

      size++;
   }

   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Object theKey)
   {
      int b = theKey.hashCode() % divisor;   // home bucket

      // search the home bucket for matching element
      // using a trailing pointer previousNode
      // first put theKey into the tailnode
      tailNode.key = (Comparable) theKey;
      HashNode currentNode = table[b],
               previousNode = null;

      while (currentNode.key.lessThan(theKey))
      {// move to next node on home bucket chain
         previousNode = currentNode;
         currentNode = currentNode.next;
      }
   
      // check if we have a match
      if (!currentNode.key.equals(theKey) || currentNode == tailNode)
         // no element with theKey in table
         return null;
   
      // we have a match, save element to be removed
      Object e = currentNode.element;
      if (previousNode != null)
         // not removing first node from home bucket chain
         previousNode.next = currentNode.next;
      else
         // remove first node from home bucket chain
         table[b] = currentNode.next;

      size--;
      return e;
   }
   
   /** output the hash table */
   public void output()
   {
      for (int i = 0; i < divisor; i++)
      {
         // output the chain table[i]
         HashNode currentNode = table[i];
         System.out.print("[");
         while(currentNode != tailNode)
         {
            if (currentNode != table[i])
               // not first element
               System.out.print(", ");
            System.out.print(currentNode.element.toString());
            currentNode = currentNode.next;
         }
         System.out.println("]");
     }

      System.out.println("Table size is " + size);
   }
   
   /** test method */
   public static void main(String [] args)
   {
      HashChainWithTail h = new HashChainWithTail(11);
      h.put(new MyInteger(80), new MyInteger(80));
      h.put(new MyInteger(40), new MyInteger(40));
      h.put(new MyInteger(65), new MyInteger(65));
      h.output();
      System.out.println();
      h.put(new MyInteger(58), new MyInteger(58));
      h.put(new MyInteger(24), new MyInteger(24));
      h.output();
      System.out.println();
      h.put(new MyInteger(2), new MyInteger(2));
      h.put(new MyInteger(13), new MyInteger(13));
      h.put(new MyInteger(46), new MyInteger(46));
      h.put(new MyInteger(16), new MyInteger(16));
      h.put(new MyInteger(7), new MyInteger(7));
      h.put(new MyInteger(21), new MyInteger(21));
      h.output();
      System.out.println();
      h.put(new MyInteger(99), new MyInteger(99));

      // update element
      h.put(new MyInteger(7), new MyInteger(29));
      h.output();
   }
}
