/** linear list as an indexed AVL tree */

package dataStructures;

import java.util.*;
import java.lang.reflect.*;
import wrappers.*;
import utilities.*;

public class LinearListAsIndexedAVLtree extends LinkedBinaryTree
                               implements LinearList
{
   // top-level nested class
   static class IAVLElement
   {
      // data members
      Object element;        // element in node
      int leftSize;          // number of elements in left subtree
      int bf;                // balance factor of node

      // constructors
      IAVLElement(int theLeftSize, int theBalanceFactor, Object theElement)
      {
         leftSize = theLeftSize;
         bf = theBalanceFactor;
         element = theElement;
      }

      IAVLElement(Object theElement)
         {element = theElement;}

      /** @return equivalent string suitable for output */
      public String toString()
         {return element.toString();}
   }

   // instance data member of LinearListAsIndexedAVLtree
   int size;  // number of elements

   // class data member used by indexOf and search
   static Object searchElement;

   // class data member used by toString
   static StringBuffer buffer;

   /** visit method that adds an element to a string buffer */
   public static void addToStringBuffer(BinaryTreeNode x)
      {buffer.append(x + ", ");}

   // define a Method variable for string conversion
   static Class [] paramType = {BinaryTreeNode.class};
                             // type of parameter for visit
   static Method theAddToStringBuffer;

   // initialize theAddToStringBuffer
   static
   { 
      try
      {
         Class lbt = LinearListAsIndexedAVLtree.class;
         theAddToStringBuffer = lbt.getMethod("addToStringBuffer", paramType);
      }
      catch (Exception e) {}
         // exception not possible
   }

   // methods of LinearList
   /** @return true iff list is empty */
   public boolean isEmpty()
       {return size == 0;}

   /** @return current number of elements in list */
   public int size()
      {return size;}

   /** @throws IndexOutOfBoundsException when
     * index is not between 0 and size - 1 */
   void checkIndex(int index)
   {
      if (index < 0 || index >= size)
         throw new IndexOutOfBoundsException
             ("index = " + index + "  size = " + size);
   }

   /** @return element with specified index
     * @throws IndexOutOfBoundsException when
     * index is not between 0 and size - 1 */
   public Object get(int index)
   {
      checkIndex(index);

      BinaryTreeNode p = root;
      while (true)
      {
         IAVLElement q = (IAVLElement) p.element;
         if (index < q.leftSize)
            // desired element is in left subtree
            p = p.leftChild;
         else
            if (index > q.leftSize)
            {// desired element is in left subtree
      	        index -= q.leftSize + 1;
                p = p.rightChild;
            }
            else 
               // found desired element
               return q.element; 
      }
   }

   /** @return index of first occurrence of theElement,
     * return -1 if elem not in list */
   public int indexOf(Object theElement)
   {
      searchElement = theElement;
      return search(root);
   }

   /** recursively search the subtree with root x
     * @return index of searchElement
     * @return -1 if searchElement not in tree */
   int search(BinaryTreeNode x)
   {
      if (x == null)
         // empty subtree
         return -1;

      // look in left subtree first
      int index = search(x.leftChild);
      if (index != -1)
         // found in left subtree
         return index;

      // compare with element in root
      if (((IAVLElement) x.element).element.equals(searchElement))
         // found the element
         return ((IAVLElement) x.element).leftSize;

      // search the right subtree
      index = search(x.rightChild);
      if (index != -1)
         // found in right subtree
         return index + ((IAVLElement) x.element).leftSize + 1;

      // not found at all
      return -1;
   }

   /** set balance factors on path from q to r */
   static void fixBF(BinaryTreeNode q, BinaryTreeNode r, int theIndex)
   {// Balance factors from q to r were originally 0.
    // They need to be changed to +1 or -1.
    // Use theIndex to find path from q to r.
   
      while (q != r)
         if (theIndex < ((IAVLElement) q.element).leftSize)
         {// height of left subtree has increased
            ((IAVLElement) q.element).bf = 1;
            q = q.leftChild;
         }
         else
         {// height of right subtree has increased
            ((IAVLElement) q.element).bf = -1;
            theIndex -= ((IAVLElement) q.element).leftSize + 1;
            q = q.rightChild;
         }
   }
   
   /** LL rotation around a, pa is parent of a and b is left child of a
     * @return true iff a is initially the left child of its parent pa */
   boolean llRotate(BinaryTreeNode pa, BinaryTreeNode a, BinaryTreeNode b)
   {
      boolean left = false;    // set to true iff a is the left child of pa
      // restructure subtree at A
      a.leftChild = b.rightChild;
      b.rightChild = a;
      ((IAVLElement) a.element).leftSize -=
                     ((IAVLElement) b.element).leftSize + 1;
      if (pa != null)
         // a is not the root
         if (a == pa.leftChild)
         {
            left = true;
            pa.leftChild = b;
         }
         else
            pa.rightChild = b;
      else
         root = b;
   
      // set balance factors
      ((IAVLElement) a.element).bf = 0;
      ((IAVLElement) b.element).bf = 0;

      return left;
   }
   
   /** RR rotation around a, pa is parent of a and b is right child of a
     * @return true iff a is initially the left child of its parent pa */
   boolean rrRotate(BinaryTreeNode pa, BinaryTreeNode a, BinaryTreeNode b)
   {
      boolean left = false;    // set to true iff a is the left child of pa
      // restructure subtree at a
      a.rightChild = b.leftChild;
      b.leftChild = a;
      ((IAVLElement) b.element).leftSize +=
                     ((IAVLElement) a.element).leftSize + 1;
      if (pa != null)
         // a is not the root
         if (a == pa.leftChild)
         {
            left = true;
            pa.leftChild = b;
         }
         else
            pa.rightChild = b;
      else
         root = b;
   
      // set balance factors
      ((IAVLElement) a.element).bf = 0;
      ((IAVLElement) b.element).bf = 0;

      return left;
   }
   
   /** LR rotation around a, pa is parent of a and b is left child of a
     * @return true iff a is initially the left child of its parent pa */
   boolean lrRotate(BinaryTreeNode pa, BinaryTreeNode a, BinaryTreeNode b)
   {
      boolean left = false;    // set to true iff a is the left child of pa
      BinaryTreeNode c = b.rightChild;
   
      // restructure subtree at a
      a.leftChild = c.rightChild;
      b.rightChild = c.leftChild;
      c.leftChild = b;
      c.rightChild = a;
      ((IAVLElement) a.element).leftSize -=
                     ((IAVLElement) b.element).leftSize +
                     ((IAVLElement) c.element).leftSize + 2;
      ((IAVLElement) c.element).leftSize +=
                     ((IAVLElement) b.element).leftSize + 1;
      if (pa != null)
         // a is not the root
         if (a == pa.leftChild)
         {
            left = true;
            pa.leftChild = c;
         }
         else 
            pa.rightChild = c;
      else 
         root = c;
   
      // set balance factors
      int bfOfC = ((IAVLElement) c.element).bf;
      if (bfOfC == 1)
      {   
         ((IAVLElement) a.element).bf = -1;
         ((IAVLElement) b.element).bf = 0;
      }
      else
         if (bfOfC == -1)
         {
            ((IAVLElement) a.element).bf = 0;
            ((IAVLElement) b.element).bf = 1;
         }
         else
         {// bfOfC = 0
            ((IAVLElement) a.element).bf = 0;
            ((IAVLElement) b.element).bf = 0;
         }
      ((IAVLElement) c.element).bf = 0;

      return left;
   }
   
   /** RL rotation around a, pa is parent of a and b is right child of a
     * @return true iff a is initially the left child of its parent pa */
   boolean rlRotate(BinaryTreeNode pa, BinaryTreeNode a, BinaryTreeNode b)
   {
      boolean left = false;    // set to true iff a is the left child of pa
      BinaryTreeNode c = b.leftChild;
   
      // restructure subtree at a
      a.rightChild = c.leftChild;
      b.leftChild = c.rightChild;
      c.leftChild = a;
      c.rightChild = b;
      ((IAVLElement) b.element).leftSize -=
                     ((IAVLElement) c.element).leftSize + 1;
      ((IAVLElement) c.element).leftSize +=
                     ((IAVLElement) a.element).leftSize + 1;
      if (pa != null)
         // a is not the root
         if (a == pa.leftChild)
         {
            left = true;
            pa.leftChild = c;
         }
         else 
            pa.rightChild = c;
      else 
         root = c;
   
      // set balance factors
      int bfOfC = ((IAVLElement) c.element).bf;
      if (bfOfC == 1)
      {   
         ((IAVLElement) a.element).bf = 0;
         ((IAVLElement) b.element).bf = -1;
      }
      else
         if (bfOfC == -1)
         {
            ((IAVLElement) a.element).bf = 1;
            ((IAVLElement) b.element).bf = 0;
         }
         else
         {// bfOfC = 0
            ((IAVLElement) a.element).bf = 0;
            ((IAVLElement) b.element).bf = 0;
         }
      ((IAVLElement) c.element).bf = 0;
     
      return left;
   }
   
   
   /** Remove the element with specified index.
     * All elements with higher index have their
     * index reduced by 1.
     * @throws IndexOutOfBoundsException when
     * when index is not between 0 and size - 1
     * @return removed element */
   public Object remove(int index)
   {
      checkIndex(index);

      size--;                // size after removal

      // define a stack to hold path taken from root
      // to physically deleted node
      ArrayStack stack = new ArrayStack();

      // p will eventually to point to node with index'th element
      BinaryTreeNode p = root; // search pointer
      IAVLElement pIAVL = (IAVLElement) p.element;
      while (index != pIAVL.leftSize)
      {
         stack.push(p);
         if (index < pIAVL.leftSize)
         {// desired element is in left subtree
            pIAVL.leftSize--;
            p = p.leftChild;
         }
         else
         {// desired element is in right subtree
      	     index -= pIAVL.leftSize + 1;
             p = p.rightChild;
         }
         pIAVL = (IAVLElement) p.element;
      }

      // save removed element
      Object elementToReturn = pIAVL.element;

      // restructure tree
      // handle case when p has two children
      if (p.leftChild != null && p.rightChild != null)
      {// p has two children, convert to 0 or 1 child case
         // find largest element in left subtree of p
         stack.push(p);
         BinaryTreeNode s = p.leftChild;
         while (s.rightChild != null)
         {// move to larger element
            stack.push(s);
            s = s.rightChild;
         }
   
         // move largest from s to p
         ((IAVLElement) p.element).element =
                ((IAVLElement) s.element).element;
         ((IAVLElement) p.element).leftSize--;
         p = s;
      }
   
      // p has at most one child
      // save child pointer in c
      BinaryTreeNode c;
      if (p.leftChild != null)
         c = p.leftChild;
      else
         c = p.rightChild;
   
      // lc will be true iff the deleted node p is
      // the left child of its parent
      boolean lc = false;

      // delete p
      if (p == root)
         root = c;
      else
         // is p a left or right child?
         if (p == ((BinaryTreeNode) stack.peek()).leftChild)
         {
            lc = true;
            ((BinaryTreeNode) stack.peek()).leftChild = c;
         }
         else
            ((BinaryTreeNode) stack.peek()).rightChild = c;

      // rebalance tree and correct balance factors
      // use stack to retrace path to root
      // set q to parent of deleted node
      BinaryTreeNode q;
      try {q = (BinaryTreeNode) stack.pop();}
      catch (Exception e)
         {return elementToReturn;}  // root was deleted
   
      while (q != null)
      {
         IAVLElement qAVL = (IAVLElement) q.element;
         if (lc)
         {// node p was deleted from the left subtree of q
            // height of left subtree reduced by 1
            qAVL.bf--;
            if (qAVL.bf == -1)
               // height of q is unchanged, nothing more to do
               return elementToReturn;
            if (qAVL.bf == -2)
            {// q is unbalanced, classify unbalance and rotate
               BinaryTreeNode b = q.rightChild,
                              pa;  // q is a node
                                   // pa is parent of a
               try {pa = (BinaryTreeNode) stack.pop();}
               catch (Exception e)
                  {pa = null;}     // stack empty
               IAVLElement bAVL = (IAVLElement) b.element;
               switch (bAVL.bf)
               {
                  case 0:           // L0 imbalance
                     lc = rrRotate(pa, q, b);
                     bAVL.bf = 1;
                     qAVL.bf = -1;  // q is a node
                     return elementToReturn;
                  case 1:           // L1 imbalance
                     lc = rlRotate(pa, q, b);
                     break;         // must continue on path to root
                  case -1:          // L-1 imbalance
                     lc = rrRotate(pa, q, b);
               }
               q = pa;
            }
            else
               // qAVL.bf is 0, back up to parent of q
               try
               {
                  BinaryTreeNode pq = (BinaryTreeNode) stack.pop();
                  if (q == pq.leftChild)
                     lc = true;
                  else
                     lc = false;
                  q = pq;
               }
               catch (Exception e)
                  {q = null;}     // stack empty
         }
         else
         {// lc is false
            // deleted from right subtree of q
            // height of right subtree reduced by 1
            qAVL.bf++;
            if (qAVL.bf == 1)
               // height of q is unchanged, nothing more to do
               return elementToReturn;
            if (qAVL.bf == 2)
            {// q is unbalanced
               // classify imbalance and rotate
               BinaryTreeNode b = q.leftChild,
                              pa;  // q is a node
                                   // pa is parent of a
               try {pa = (BinaryTreeNode) stack.pop();}
               catch (Exception e)
                  {pa = null;}     // stack empty
               IAVLElement bAVL = (IAVLElement) b.element;
               switch (bAVL.bf)
               {
                  case 0:                // R0 imbalance
                     lc = llRotate(pa, q, b);
                     bAVL.bf = -1;
                     qAVL.bf = 1;        // q is a node
                     return elementToReturn;
                  case 1:                // R1 imbalance
                     lc = llRotate(pa, q, b);
                     break;              // must continue on path to root
                  case -1:               // R-1 imbalance
                     lc = lrRotate(pa, q, b);
               }
               q = pa;
            }
            else
               // qAVL.bf is 0, back up to parent of q
               try
               {
                  BinaryTreeNode pq = (BinaryTreeNode) stack.pop();
                  if (q == pq.leftChild)
                     lc = true;
                  else
                     lc = false;
                  q = pq;
               }
               catch (Exception e)
                  {q = null;}     // stack empty
         }   
      }     
      return elementToReturn;
   }
   
   /** Insert an element with specified index.
     * All elements with equal or higher index
     * have their index increased by 1.
     * @throws IndexOutOfBoundsException when
     * index is not between 0 and size */
   public void add(int index, Object obj)
   {
      if (index < 0 || index > size)
         // invalid list position
         throw new IndexOutOfBoundsException
             ("index = " + index + "  size = " + size);

      BinaryTreeNode p = root,     // search pointer
                     pp = null,    // parent of p
                     a = null,     // node with bf != 0
                     pa = null;    // parent of a
      int indexAta = index;        // value of index when at node a

      // find place to insert theElement
      // also record most recent node with bf != 0
      // in a and its parent in pa
      boolean leftMove = true;              // was last move to a left child?
      while (p != null)
      {// examine p.element.leftSize
         IAVLElement pElement = (IAVLElement) p.element;
         if (pElement.bf != 0)
         {// new candidate for a node
            a = p;
            pa = pp;
            indexAta = index;
         }
         pp = p;
         // move p to a child
         if (index <= pElement.leftSize)
         {// the insert will add a new node in left subtree
            pElement.leftSize++;
            p = p.leftChild;
            leftMove = true;
         }
         else
         {
            index -= pElement.leftSize + 1;
            p = p.rightChild;
            leftMove = false;
         }
      }
   
      // get a node for theElement and attach to pp
      IAVLElement q = new IAVLElement(0, 0, obj);
      BinaryTreeNode r = new BinaryTreeNode(q);
      size++;
      if (root != null)
         // the tree is not empty
         if (leftMove)
            pp.leftChild = r;
         else
            pp.rightChild = r;
      else
      {// insertion into empty tree
         root = r;
         return;
      }
      
      // see if we must rebalance or simply change balance factors
      index = indexAta;
      if (a != null)
      {// possible rebalancing needed
         IAVLElement aElement = (IAVLElement) a.element;
         if (aElement.bf < 0)
            // bf = -1 before insertion
            if (index < aElement.leftSize)
            {// insertion in left subtree
               // height of left subtree has increased by 1
               // new bf of a is 0, no rebalancing
               aElement.bf = 0;
               // fix bf on path from a to r
               fixBF(a.leftChild, r, indexAta);
            }
            else
            {// insertion in right subtree
               // bf of a is -2, rebalance
               BinaryTreeNode b = a.rightChild;
               index -= aElement.leftSize + 1;  // index of r in b
               if (index > ((IAVLElement) b.element).leftSize)
               {// RR case
                  index -= ((IAVLElement) b.element).leftSize + 1;
                  fixBF(b.rightChild, r, index);
                  rrRotate(pa, a, b);
               }
               else
               {// RL case
                  fixBF(b.leftChild, r, index);
                  rlRotate(pa, a, b);
               }
            }
         else // bf = +1 before insertion
            if (index > aElement.leftSize)
            {// insertion in right subtree
               // height of right subtree has increased by 1
               // new bf of a is 0, no rebalancing
               aElement.bf = 0;
               // fix bf on path from a to r
               index -= aElement.leftSize + 1;
               fixBF(a.rightChild, r, index);
            }
            else
            {// insertion in left subtree
               // bf of a is +2, rebalance
               BinaryTreeNode b = a.leftChild;
               if (index < ((IAVLElement) b.element).leftSize)
               {// LL case
                  fixBF(b.leftChild, r, index);
                  llRotate(pa, a, b);
               }
               else
               {// LR case
                  index -= ((IAVLElement) b.element).leftSize + 1;
                  fixBF(b.rightChild, r, index);
                  lrRotate(pa, a, b);
               }
            }
      }
      else // a is nulL, no rebalancing
         fixBF(root, r, index);
   }

   /** convert to a string */
   public String toString()
   {
      if (size == 0)
         return "[]";

      buffer = new StringBuffer("["); 
      inOrder(theAddToStringBuffer);

      // remove last blank character
      buffer.setLength(buffer.length() - 1);

      // overwrite last ,
      buffer.setCharAt(buffer.length() - 1, ']');

      // create equivalent String
      return new String(buffer);
   }

   /** test program */
   public static void main(String [] args)
   {
      // test default constructor
      LinearListAsIndexedAVLtree x = new LinearListAsIndexedAVLtree();

      // test size
      System.out.println("Initial size is " + x.size());

      // test isEmpty
      if (x.isEmpty())
         System.out.println("The list is empty");
      else System.out.println("The list is not empty");

      // test add
      x.add(0, new Integer(2));
      x.add(1, new Integer(6));
      x.add(0, new Integer(1));
      x.add(2, new Integer(4));
      System.out.println("List size is " + x.size());

      // test toString
      System.out.println("The list is " + x);

      // test indexOf
      int index = x.indexOf(new Integer(4));
      if (index < 0)
         System.out.println("4 not found");
      else System.out.println("The index of 4 is " + index);

      index = x.indexOf(new Integer(3));
      if (index < 0)
         System.out.println("3 not found");
      else System.out.println("The index of 3 is " + index);

      // test get
      System.out.println("Element at 0 is " + x.get(0));
      System.out.println("Element at 3 is " + x.get(3));

      // test remove
      System.out.println("Removed " + x.remove(1) + " The list is " + x);
      System.out.println("Removed " + x.remove(2) + " The list is " + x);

      if (x.isEmpty())
         System.out.println("The list is empty");
      else System.out.println("The list is not empty");

      System.out.println("List size is " + x.size());
      System.out.println();
      System.out.println("Starting new list");

      LinearListAsIndexedAVLtree y = new LinearListAsIndexedAVLtree();

      // test insert 
      y.add(0, new Integer(21));
      y.add(1, new Integer(26));
      y.add(2, new Integer(30));
      System.out.println("An RR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // test LL rotation
      y.add(0, new Integer(9));
      y.add(0, new Integer(4));
      System.out.println("An LL rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // test LR rotation
      y.add(2, new Integer(14));
      System.out.println("An LR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // test RL rotation
      y.add(5, new Integer(28));
      System.out.println("An RL rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // another RL rotation
      y.add(3, new Integer(18));
      y.add(3, new Integer(15));
      System.out.println("An RL rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // yet another RLrotation
      y.add(2, new Integer(10));
      System.out.println("An RL rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // an LL rotation
      y.add(0, new Integer(2));
      System.out.println("An LL rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // an LR rotation
      y.add(1, new Integer(3));
      System.out.println("An LR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // another LR rotation
      y.add(3, new Integer(7));
      System.out.println("An LR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // test get
      for (int i = 0; i < 13; i ++)
         System.out.println("Element " + i + " is " + y.get(i));
      System.out.println();

      // now test the remove
      // L0/RR rotation
      y.remove(0);
      y.remove(0);
      System.out.println("An L0/RR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // R-1/RR rotation
      y.remove(3);
      System.out.println("An R-1/LR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // L-1/RR rotation
      y.remove(5);
      y.remove(0);
      y.remove(1);
      System.out.println("An L-1/RR rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
      System.out.println();
   
      // L1/RL rotation
      y.remove(6);
      y.remove(1);
      y.remove(0);
      y.remove(0);
      System.out.println("An L1/RL rotation has been done");
      System.out.println("List is " + y);
      System.out.println("Elements in postorder order are");
      y.postOrderOutput();
      System.out.println();
      System.out.println("List size is " + y.size());
   }
}
