/** linked binary trees with a postorder iterator */

package dataStructures;

import java.util.*;

public class LinkedBinaryTreeWithPostorderIterator extends LinkedBinaryTree
{
   // define a class for use by the iterator
   static class StackElement
   {
      // data members
      BinaryTreeNode node;
      boolean leftChild;   // true iff we move to the left child of node

      // constructor
      public StackElement(BinaryTreeNode theNode, boolean theLeftChild)
      {
         node = theNode;
         leftChild = theLeftChild;
      }
   }

   /** create and return an iterator */
   public Iterator iterator()
      {return new PostorderIterator();}

   /** postorder iterator */
   private class PostorderIterator implements Iterator
   {
      // data members
      private BinaryTreeNode nextNode;  // next node in postorder
      private ArrayStack stack = new ArrayStack(10);
   
      // constructor
      /** set nextNode to first node in postorder */
      public PostorderIterator()
         {nextNode = (root == null) ? null : leftMostLeaf(root);}
   
      // methods
      /** @return the leftmost leaf in the subtree with root t
        * save nodes on the path to this leaf on the stack */
      BinaryTreeNode leftMostLeaf(BinaryTreeNode t)
      {
         BinaryTreeNode currentNode = t;
         while (true)
         {
            // move to leftmost node in subtree
            while (currentNode.leftChild != null)
            {
               stack.push(new StackElement(currentNode, true));
               currentNode = currentNode.leftChild;
            }

            // is this node a leaf?
            if (currentNode.rightChild == null)
               // it is a leaf, must be leftmost leaf
               return currentNode;

            // leftmost leaf is in right subtree of currentNode
            stack.push(new StackElement(currentNode, false));
            currentNode = currentNode.rightChild;
         }
      }

      /** @return true iff tree has more iterator */
      public boolean hasNext()
         {return nextNode != null;}

      /** @return next element in postorder
        * @throws NoSuchElementException
        * when there is no next element */
      public Object next()
      {
         if (nextNode != null)
         {
            Object obj = nextNode.element;
            
            // determine next postorder node
            if (stack.empty())
            {// there is no next node
               nextNode = null;
            }
            else
            {// there is a next node
               StackElement sElement = (StackElement) stack.pop();
               nextNode = sElement.node;
               if (sElement.leftChild && nextNode.rightChild != null)
               {// returning from left subtree of nextNode
                // incorrect nextNode, correct next node is leftmost
                // leaf in right subtree of nextNode
                  stack.push(new StackElement(nextNode, false));
                  nextNode = leftMostLeaf(nextNode.rightChild);
               }
            }
            return obj;
         }
         else throw new NoSuchElementException("No next element");
      }

      /** unsupported method */
      public void remove()
      {
         throw new UnsupportedOperationException
                   ("remove not supported");
      }   
   }

   /** test program */
   public static void main(String [] args)
   {
      LinkedBinaryTreeWithPostorderIterator
                       a = new LinkedBinaryTreeWithPostorderIterator(),
                       w = new LinkedBinaryTreeWithPostorderIterator(), 
                       x = new LinkedBinaryTreeWithPostorderIterator(), 
                       y = new LinkedBinaryTreeWithPostorderIterator(), 
                       z = new LinkedBinaryTreeWithPostorderIterator();
      y.makeTree(new Integer(1), a, a);
      z.makeTree(new Integer(2), a, a);
      x.makeTree(new Integer(3), y, z);
      y.makeTree(new Integer(4), x, a);
      w.makeTree(new Integer(5), a, a);
      z.makeTree(new Integer(6), a, a);
      x.makeTree(new Integer(7), w, z);
      y.makeTree(new Integer(8), x, y);

      System.out.println("Postorder sequence is ");
      y.postOrderOutput();
      System.out.println();

      System.out.println("Postorder iterator yields");
      Iterator iy = y.iterator();
      while (iy.hasNext())
         System.out.print(iy.next() + " ");
      System.out.println();

      // try another tree
      w.makeTree(new Integer(9), a, y);
      System.out.println("Postorder sequence is ");
      w.postOrderOutput();
      System.out.println();

      System.out.println("Postorder iterator yields");
      Iterator iw = w.iterator();
      while (iw.hasNext())
         System.out.print(iw.next() + " ");
      System.out.println();
   }
}
