/** a chain with elements in decreasing order, multiple elements with the
  * same value are permitted */

package dataStructures;

import java.util.*;

public class SortedChainWithNoKey
{
   // data members
   protected ChainNode firstNode;
   protected int size;

   // constructors
   /** create an empty sorted chain */
   public SortedChainWithNoKey(int initialCapacity)
   {
       // the default initial values of firstNode and size
       // are null and 0, respectively
   }

   public SortedChainWithNoKey()
      {this(0);}

   // methods
   /** @return true iff the chain is empty */
   public boolean isEmpty()
       {return size == 0;}

   /** @return current number of elements in list */
   public int size()
      {return size;}

   /** @return element with specified value
     * @return null if there is no matching element */
   public Object get(Comparable theValue)
   {
      ChainNode currentNode = firstNode;

      // search for match with theKey
      while (currentNode != null && 
             theValue.compareTo(currentNode.element) < 0)
         currentNode = currentNode.next;

      // verify match
      if (currentNode != null && theValue.equals(currentNode.element))
         // yes, found match
         return currentNode.element;

      // no match
      return null;
   }
   
   /** insert an element, multiple elements with the same value
     * are permitted */
   public void put(Comparable theElement)
   {
      ChainNode p = firstNode,
                tp = null; // tp trails p
   
      // move tp so that theElement can be inserted after tp
      while (p != null && theElement.compareTo(p.element) < 0)
      {
         tp = p;
         p = p.next;
      }
   
      // set up node for theElement
      ChainNode q = new ChainNode(theElement, p);
   
      // insert node just after tp
      if (tp == null)
         firstNode = q;
      else
         tp.next = q;

      size++;
   }

   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Comparable theValue)
   {
      ChainNode p = firstNode,
                tp = null; // tp trails p
      
      // search for match with theValue
      while (p != null && theValue.compareTo(p.element) < 0)
      {
         tp = p;
         p = p.next;
      }
   
      // verify match
      if (p != null && theValue.equals(p.element)) 
      {// found a match
         Object e = p.element;  // the matching element
   
         // remove p from the chain
         if (tp == null)
            firstNode = p.next;  // p is first node
         else
            tp.next = p.next;
   
         size--;
         return e;
      }
   
      // no matching element to remove
      return null;
   }
   
   /** convert to a string */
   public String toString()
   {
      StringBuffer s = new StringBuffer("["); 
      ChainNode currentNode = firstNode;
      while (currentNode != null)
      {
         if (currentNode != firstNode)
            // not first element
            s.append(", ");
         s.append(currentNode.element.toString());
         currentNode = currentNode.next;
      }
      s.append("]");

      // create equivalent String
      return new String(s);
   }
   

   /** create and return an iterator */
   public Iterator iterator()
      {return new ChainIterator();}

   /** chain iterator */
   private class ChainIterator implements Iterator
   {
      // data member
      private ChainNode nextNode;
   
      // constructor
      public ChainIterator()
         {nextNode = firstNode;}
   
      // methods
      /** @return true iff list has a next element */
      public boolean hasNext()
         {return nextNode != null;}
   
      /** @return next element in list
        * @throws NoSuchElementException
        * when there is no next element */
      public Object next()
      {
         if (nextNode != null)
         {
            Object elementToReturn = nextNode.element;
            nextNode = nextNode.next;
            return elementToReturn;
         }
         else
            throw new NoSuchElementException("No next element");
      }

      /** unsupported method */
      public void remove()
      {
         throw new UnsupportedOperationException
                   ("remove not supported");
      }   
   }
   
   /** test program */
   public static void main(String [] args)
   {
      // test default constructor
      SortedChainWithNoKey x = new SortedChainWithNoKey();

      // test put
      x.put(new Integer(2));
      System.out.println("The list is " + x);
      x.put(new Integer(6));
      System.out.println("The list is " + x);
      x.put(new Integer(1));
      System.out.println("The list is " + x);
      x.put(new Integer(4));
      System.out.println("The list is " + x);
      x.put(new Integer(6));
      System.out.println("The list is " + x);
      System.out.println("The list size is " + x.size());

      // test get
      System.out.println("element " + x.get(new Integer(2))
                         + " has value 2");
      System.out.println("element " + x.get(new Integer(1))
                         + " has value 1");
      System.out.println("element " + x.get(new Integer(6))
                         + " has value 6");

      // test remove
      System.out.println("removed element " + x.remove(new Integer(2))
                         + " with value 2");
      System.out.println("The list is " + x);
      System.out.println("removed element " + x.remove(new Integer(1))
                         + " with value 1");
      System.out.println("The list is " + x);
      System.out.println("removed element " + x.remove(new Integer(6))
                         + " with value 6");
      System.out.println("The list is " + x);
      System.out.println("The list size is " + x.size());
   }
}
