
package dataStructures;

import java.util.*;
import wrappers.*;
import utilities.*;

public class SortedList implements Dictionary
{
   // top-level nested class
   protected static class SortedListNode
   {
      // data members
      protected Comparable key;
      protected Object element;

      // constructors
      protected SortedListNode() {}
     
      protected SortedListNode(Object theKey, Object theElement)
      {
         key = (Comparable) theKey;
         element = theElement;
      }
   }

   // data members of SortedList
   protected SortedListNode [] element;  // array of elements
   protected int size;                   // number of elements in array

   // constructors
   /** create a list with initial capacity initialCapacity
     * @exception IllegalArgumentException thrown
     * if initialCapacity < 1 */
   public SortedList(int initialCapacity)
   {
      if (initialCapacity < 1)
         throw new IllegalArgumentException
            ("SortedList.constructor: " +
             "initialCapacity must be >= 1");
      // size has the default initial value of 0
      element = new SortedListNode [initialCapacity];
   }

   /** create a list with initial capacity 10 */
   public SortedList()
   {// use default capacity of 10
      this(10);
   }

   // methods
   /** @return true iff list is empty */
   public boolean isEmpty()
       {return size == 0;}

   /** @return current number of elements in list */
   public int size()
      {return size;}

   /** @return element with specified key
     * @return null if there is no matching element */
   public Object get(Object theKey)
   {
      // search from left to right for matching element
      int i = 0;
      while (i < size && element[i].key.lessThan(theKey))
         i++;
   
      // verify match
      if (i >= size || !element[i].key.equals(theKey))
         // no matching element
         return null;
      else
         // there is a match
         return element[i].element;
   }
   
   /** insert an element with the specified key
     * overwrite old element if there is already an
     * element with the given key */
   public void put(Object theKey, Object theElement)
   {
      // examine elements from left to right
      int i = 0;
      while (i < size && element[i].key.lessThan(theKey))
         i++;
   
      // verify match
      if (i < size && element[i].key.equals(theKey))
      {// replace old element
         element[i].element = theElement;
         return;
      }

      if (size == element.length)
         // no space for new element, double capacity
         element = (SortedListNode []) ChangeArraySize.changeSize1D
                      (element, 2 * element.length);
   
      // create space for new element by moving
      // [i:size - 1] one spot right
      for (int j = size-1; j >= i; j--)
         element[j + 1] = element[j];

      // insert at element[i]
      element[i] = new SortedListNode(theKey, theElement);
      size++;
   }

   /** @return matching element and remove it
     * @return null if no matching element */
   public Object remove(Object theKey)
   {
      // search for match with theKey
      int i = 0;
      while (i < size && element[i].key.lessThan(theKey))
         i++;
   
      // verify match
      if (i >= size || !element[i].key.equals(theKey))
         // no matching element
         return null;
   
      // save matching element in e and remove from list
      Object e = element[i].element;
      for (int j = i + 1; j < size; j++)
         // shift element left by one
         element[j-1] = element[j];
      size--;  // new list size

      return e;
   }
   
   /** convert to a string */
   public String toString()
   {
      StringBuffer s = new StringBuffer("["); 
      if (size != 0)
      {// nonempty list
         // do first element
         s.append(element[0].element.toString());
         // do remaining elements
         for (int i = 1; i < size; i++)
            s.append(", " + element[i].element.toString());
      }
      s.append("]");

      // create equivalent String
      return new String(s);
   }

   /** create and return an enumerator */
   public Enumeration elements()
   {
      return new Enumerator();
   }

   private class Enumerator implements Enumeration
   {
      // data members
      private int nextIndex;  // index of next element
   
      // constructor
      public Enumerator()
      {
         // nextIndex has the default initial value 0
      }
   
      // methods
      /** @return true iff list has more elements */
      public boolean hasMoreElements()
      {
         return nextIndex < size;
      }
   
      /** @return next element in list
        * @exception NoSuchElementException
        * thrown if there is no next element */
      public Object nextElement()
      {
         if (nextIndex < size)
            return element[nextIndex++];
         else throw new NoSuchElementException
                 ("SortedList."
                   + "nextElement: No next element");
      }
   }
   
   /** test program */
   public static void main(String [] args)
   {
      // test default constructor
      SortedList x = new SortedList();

      // test put
      x.put(new MyInteger(2), new MyInteger(12));
      System.out.println("The list is " + x);
      x.put(new MyInteger(6), new MyInteger(16));
      System.out.println("The list is " + x);
      x.put(new MyInteger(1), new MyInteger(11));
      System.out.println("The list is " + x);
      x.put(new MyInteger(4), new MyInteger(14));
      System.out.println("The list is " + x);
      x.put(new MyInteger(6), new MyInteger(26));
      System.out.println("The list is " + x);

      // test get
      System.out.println("element " + x.get(new MyInteger(2))
                         + " has key 2");
      System.out.println("element " + x.get(new MyInteger(1))
                         + " has key 1");
      System.out.println("element " + x.get(new MyInteger(6))
                         + " has key 6");

      // test remove
      System.out.println("removed element " + x.remove(new MyInteger(2))
                         + " with key 2");
      System.out.println("The list is " + x);
      System.out.println("removed element " + x.remove(new MyInteger(1))
                         + " with key 1");
      System.out.println("The list is " + x);
      System.out.println("removed element " + x.remove(new MyInteger(6))
                         + " with key 6");
      System.out.println("The list is " + x);
   }
}
