
/** Toeplitz matrix mapped into a one-dimensional array */

package dataStructures;

import utilities.*;
import wrappers.*;

public class ToeplitzMatrix
{
   // data members
   int rows;            // matrix dimension
   Object zero;         // zero element
   Object [] element;   // element array

   // constructor
   public ToeplitzMatrix(int theRows, Object theZero)
   {
      // validate theRows
      if (theRows < 1)
         throw new IllegalArgumentException
                   ("number of rows must be > 0");
   
      // create and initialize the matrix
      rows = theRows;
      zero = theZero;
      element = new Object [2 * rows - 1];
      for (int i = 0; i <  2 * rows - 1; i++)
         element[i] = zero;
   }
   
   // methods
   /** @throws IndexOutOfBoundsException when i < 1
     * or j < 1 or i > rows or j > rows */
   void checkIndex(int i, int j)
   {
      if (i < 1 || j < 1 || i > rows || j > rows)
         throw new IndexOutOfBoundsException
                   ("i = " + i + " j = " + j +
                    " rows = " + rows + " cols = " + rows);
   }

   /** @return the element this(i,j)
     * @throws IndexOutOfBoundsException when i or j invalid */
   public Object get(int i, int j)
   {
      checkIndex(i, j);
      return element[i - j + rows - 1];
   }
   
   /** set this(i,j) = newValue
     * @throws IndexOutOfBoundsException when i or j invalid */
   public void set(int i, int j, Object newValue)
   {
      checkIndex(i, j);

      // store newValue
      element[i - j + rows - 1] = newValue;
   }
   
   /** @return the product of this and matrix b */
   public Computable[][] multiply(ToeplitzMatrix b)
   {
      // make sure both matrices are of the same size
      if (rows != b.rows)
         throw new IllegalArgumentException
               ("matrices must have the same size");
   
      // create array for result
      Computable [][] c = new Computable[rows + 1][rows + 1];

      // compute the product of this and b
      for (int i = 1; i <= rows; i++)
         for (int j = 1; j <= rows; j++)
         {// compute c[i][j]
            Computable sum = (Computable) ((CloneableObject) zero).clone();
            for (int k = 1; k <= rows; k++)
               sum.increment(((Computable) element[i - k + rows - 1])
                             .multiply(b.element[k - j + rows - 1]));
            c[i][j] = sum;
          }

      return c;
   }
   
   /** test program */
   public static void main(String [] args)
   {
      // create two matrices
      int n = 3;
      ToeplitzMatrix x = new ToeplitzMatrix(n, new MyInteger(0));
      ToeplitzMatrix y = new ToeplitzMatrix(n, new MyInteger(0));

      // set a few elements in x
      x.set(1, 1, new MyInteger(1));
      x.set(1, 2, new MyInteger(2));
      x.set(1, 3, new MyInteger(0));
      x.set(2, 1, new MyInteger(3));

      // set a few elements in y
      y.set(1, 1, new MyInteger(2));
      y.set(1, 2, new MyInteger(1));
      y.set(1, 3, new MyInteger(1));
      y.set(3, 1, new MyInteger(2));

      // retrieve a few elements
      System.out.println(x.get(1, 3));
      System.out.println(x.get(2, 2));
      System.out.println(x.get(3, 2));
      System.out.println(x.get(2, 3));

      // multiply and output
      Computable [][] c = x.multiply(y);
      System.out.println("The product is");
      for (int i = 1; i <= n; i++)
      {
         for (int j = 1; j <= n; j++)
            System.out.print(c[i][j] + " ");
         System.out.println();
      }
   }
}
